#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.dispatch import receiver

from ..dataformats.models import DataFormat

from .models import Algorithm, AlgorithmEndpoint


#----------------------------------------------------------


# These two auto-delete files from filesystem when they are unneeded:
@receiver(models.signals.post_delete, sender=Algorithm)
def auto_delete_file_on_delete(sender, instance, **kwargs):
    """Deletes file from filesystem when ``Algorithm`` object is deleted.
    """
    if instance.declaration_file:
        instance.declaration_file.delete(save=False)

    if instance.description_file:
        instance.description_file.delete(save=False)

    if instance.source_code_file:
        instance.source_code_file.delete(save=False)


@receiver(models.signals.pre_save, sender=Algorithm)
def auto_delete_file_on_change(sender, instance, **kwargs):
    """Deletes file from filesystem when ``Algorithm`` object is changed."""
    if not instance.pk:
        return False

    try:
        old_file = Algorithm.objects.get(pk=instance.pk).declaration_file
    except Algorithm.DoesNotExist:
        return False

    if old_file != instance.declaration_file:
        old_file.delete(save=False)

    try:
        old_descr = Algorithm.objects.get(pk=instance.pk).description_file
    except Algorithm.DoesNotExist:
        return False

    if old_descr != instance.description_file:
        old_descr.delete(save=False)

    try:
        old_code = Algorithm.objects.get(pk=instance.pk).source_code_file
    except Algorithm.DoesNotExist:
        return False

    if old_code != instance.source_code_file:
        old_code.delete(save=False)


#----------------------------------------------------------


@receiver(models.signals.pre_delete, sender=Algorithm)
def delete_endpoints(sender, instance, **kwargs):
    instance.endpoints.all().delete()


#----------------------------------------------------------


@receiver(models.signals.post_save, sender=Algorithm)
def create_endpoints(sender, instance, **kwargs):
    """Creates the endpoints required by the algorithm, if they don't exist"""

    wrapper = instance.validate(instance.declaration)

    used_endpoints = []
    used_names = set()

    def _create_or_reuse_endpoint(isinput, name, formatname, channel):

        (df_author, df_name, df_version) = formatname.split('/')
        df = DataFormat.objects.get(
                author__username=df_author,
                name=df_name,
                version=df_version,
                )

        endpoint = AlgorithmEndpoint.objects.filter(
                algorithm = instance,
                input = isinput,
                name = name,
                #channel = channel_name,
                dataformat = df,
            )

        if not endpoint:
            endpoint            = AlgorithmEndpoint()
            endpoint.algorithm  = instance
            endpoint.input      = isinput
            endpoint.name       = name
            endpoint.channel    = channel
            endpoint.dataformat = df
            endpoint.save()
            used_endpoints.append(endpoint)
        else:
            endpoint[0].channel = channel
            endpoint[0].save() # reset channel name with new standard
            used_endpoints.append(endpoint[0])

    for group in wrapper.groups:
        group_name = group.get('name', '__io_group_%d__' % len(used_names))
        used_names.add(group_name)
        for name, details in group['inputs'].items():
            _create_or_reuse_endpoint(True, name, details['type'], group_name)
        for name, details in group.get('outputs', {}).items():
            _create_or_reuse_endpoint(False, name, details['type'], group_name)

    # cleans up unused endpoints
    all_endpoints = AlgorithmEndpoint.objects.filter(algorithm=instance)
    for endpoint in all_endpoints:
        if endpoint not in used_endpoints:
            endpoint.delete()
