/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 * 
 * This file is part of the beat.web module of the BEAT platform.
 * 
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 * 
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.algorithms.models.js

   Declaration of the algorithms-related classes
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.algorithms === undefined)
    beat.algorithms = {}

if (beat.algorithms.models === undefined)
    beat.algorithms.models = {}


/*********************************** CLASS: Dataset *************************************/


//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.algorithms.models.Algorithm = function(data, duplicate_of)
{
    // Attributes
    this.name              = data.name;
    this.short_description = data.short_description;
    this.hash              = data.hash;
    this.inputs            = data.inputs;
    this.outputs           = data.outputs;
    this.parameters        = data.parameters;
    this.splittable        = data.splittable;
    this.valid             = data.valid;
    this.language          = data.language;
    this.duplicate_of      = (duplicate_of !== undefined ? duplicate_of : null);
}


//----------------------------------------------------------------------------------------
// Retrieves the details about an input of the algorithm
//----------------------------------------------------------------------------------------
beat.algorithms.models.Algorithm.prototype.input = function(name)
{
    for (var i = 0; i < this.inputs.length; ++i)
    {
        if (this.inputs[i].name == name)
            return this.inputs[i];
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieves the details about an output of the algorithm
//----------------------------------------------------------------------------------------
beat.algorithms.models.Algorithm.prototype.output = function(name)
{
    for (var i = 0; i < this.outputs.length; ++i)
    {
        if (this.outputs[i].name == name)
            return this.outputs[i];
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieves a list of all the channels of the inputs
//----------------------------------------------------------------------------------------
beat.algorithms.models.Algorithm.prototype.inputChannels = function()
{
    var channels = [];

    for (var i = 0; i < this.inputs.length; i++)
    {
        if (channels.indexOf(this.inputs[i].channel) == -1)
            channels.push(this.inputs[i].channel);
    }

    return channels;
}


//----------------------------------------------------------------------------------------
// Retrieves the inputs belonging to the specified group
//----------------------------------------------------------------------------------------
beat.algorithms.models.Algorithm.prototype.inputsOfChannel = function(channel)
{
    return this.inputs.filter(function(input) {
        return input.channel == channel;
    });
}


//----------------------------------------------------------------------------------------
// Retrieves a list of all the channels of the outputs
//----------------------------------------------------------------------------------------
beat.algorithms.models.Algorithm.prototype.outputChannels = function()
{
    var channels = [];

    for (var i = 0; i < this.outputs.length; i++)
    {
        if (channels.indexOf(this.outputs[i].channel) == -1)
            channels.push(this.outputs[i].channel);
    }

    return channels;
}


//----------------------------------------------------------------------------------------
// Retrieves the outputs belonging to the specified group
//----------------------------------------------------------------------------------------
beat.algorithms.models.Algorithm.prototype.outputsOfChannel = function(channel)
{
    return this.outputs.filter(function(output) {
        return output.channel == channel;
    });
}



/********************************* CLASS: DatasetsList **********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Allows to easily manipulate a list of algorithms
//
// declaration: JSON declaration of the list, as returned by the Web API
//----------------------------------------------------------------------------------------
beat.algorithms.models.AlgorithmsList = function(declaration, valid_only)
{
    if (valid_only === undefined)
        valid_only = false;

    // Attributes
    this.data = new Array();

    // Initialisation
    var shared_hashes = declaration.filter(function(algorithm) {
            return (algorithm.name.indexOf('/private/') == -1);
        }).map(function(algorithm) {
            return algorithm.hash;
        });

    for (var i = 0; i < declaration.length; ++i)
    {
        var algorithm = declaration[i];

        if (valid_only && !algorithm.valid)
            continue;

        if ((algorithm.name.indexOf('/private/') == -1) || (shared_hashes.indexOf(algorithm.hash) == -1))
        {
            this.data.push(new beat.algorithms.models.Algorithm(algorithm));
        }
        else
        {
            for (var j = 0; j < declaration.length; ++j)
            {
                var algorithm2 = declaration[j];

                if (algorithm.hash == algorithm2.hash)
                {
                    this.data.push(new beat.algorithms.models.Algorithm(algorithm, algorithm2.name));
                    break;
                }
            }
        }
    }

    this.length = this.data.length;
}


//----------------------------------------------------------------------------------------
// Retrieve an algorithm by name
//----------------------------------------------------------------------------------------
beat.algorithms.models.AlgorithmsList.prototype.get = function(name)
{
    for (var i = 0; i < this.data.length; i++)
    {
        var algorithm = this.data[i];

        if (algorithm.name == name)
            return algorithm;
    }

    return null;
}


//----------------------------------------------------------------------------------------
// Retrieve an algorithm by index
//----------------------------------------------------------------------------------------
beat.algorithms.models.AlgorithmsList.prototype.getByIndex = function(index)
{
    if (index < this.data.length)
        return this.data[index];

    return null;
}


//----------------------------------------------------------------------------------------
// Returns an iterator over the algorithms
//----------------------------------------------------------------------------------------
beat.algorithms.models.AlgorithmsList.prototype.iterator = function()
{
    return new beat.algorithms.models.AlgorithmsIterator(this);
}



/******************************* CLASS: DatasetsIterator ********************************/

//----------------------------------------------------------------------------------------
// Constructor
//
// Allows to easily iterate through a list of algorithms
//----------------------------------------------------------------------------------------
beat.algorithms.models.AlgorithmsIterator = function(algorithms)
{
    // Attributes
    this.algorithms = algorithms;
    this.next_index = 0;
}


//----------------------------------------------------------------------------------------
// Indicates if the iteration can continue
//----------------------------------------------------------------------------------------
beat.algorithms.models.AlgorithmsIterator.prototype.hasMoreData = function()
{
    return (this.next_index < this.algorithms.length);
}


//----------------------------------------------------------------------------------------
// Returns the next algorithm
//----------------------------------------------------------------------------------------
beat.algorithms.models.AlgorithmsIterator.prototype.next = function()
{
    var algorithm = this.algorithms.getByIndex(this.next_index);

    this.next_index += 1;

    return algorithm;
}
