#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

import simplejson as json
from datetime import datetime

from django.contrib.auth.models import User
from django.conf import settings
from django.core.urlresolvers import reverse

from .models import Database

from ..dataformats.models import DataFormat

from ..common.testutils import BaseTestCase

class DatabaseAPIBase(BaseTestCase):
    DATABASE = {
            "root_folder": "/path/to/root/folder",
            "protocols": [
                {
                    "name": "protocol1",
                    "template": "test",
                    "sets": [
                        {
                            "name": "set1",
                            "template": "set",
                            "view": "dummy",
                            "outputs": {
                                "out": settings.SYSTEM_ACCOUNT + "/float/1"
                                }
                            }
                        ]
                    }
                ]
            }


    def setUp(self):
        # Users
        self.system_user = User.objects.create_user(settings.SYSTEM_ACCOUNT,
                                                    settings.SYSTEM_ACCOUNT + '@test.org', '1234')
        self.system_user.is_staff = True
        self.system_user.save()
        user = User.objects.create_user('jackdoe', 'jackdoe@test.org', '1234')
        User.objects.create_user('johndoe', 'johndoe@test.org', '1234')


    def tearDown(self):
        pass


class AttestationCreationAPI(DatabaseAPIBase):
    def setUp(self):
        super(AttestationCreationAPI, self).setUp()

        self.url = reverse('api_databases:all')

    def test_no_access_for_anonymous_user(self):
        response = self.client.post(self.url)
        self.checkResponse(response, 403)


    def test_no_access_for_non_admin_user(self):
        self.client.login(username='jackdoe', password='1234')
        response = self.client.post(self.url)
        self.checkResponse(response, 403)


    def test_create_database_failure(self):
        self.client.login(username=settings.SYSTEM_ACCOUNT, password='1234')
        db_name = 'test_db'
        response = self.client.post(self.url,
            json.dumps({
                'name': db_name,
                'declaration': self.DATABASE
            }), content_type='application/json')

        data = self.checkResponse(response, 400)


    def test_create_database(self):
        # Create a dataformat
        (dataformat, errors) = DataFormat.objects.create_dataformat(self.system_user, 'float', '')
        assert dataformat, errors
        dataformat.share()

        self.client.login(username=settings.SYSTEM_ACCOUNT, password='1234')
        db_name = 'test_db'
        response = self.client.post(self.url,
            json.dumps({
                'name': db_name,
                'declaration': self.DATABASE
            }), content_type='application/json')

        data = self.checkResponse(response, 201, content_type='application/json')

        self.assertTrue(data['name'] == db_name)

        databases = Database.objects.all()
        self.assertEqual(databases.count(), 1)
