#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from rest_framework import serializers

from ..common.serializers import ContributionSerializer, VersionableSerializer, DynamicFieldsSerializer, ContributionCreationSerializer
from ..algorithms.models import Algorithm
from ..databases.models import Database, DatabaseSet

from .models import DataFormat

import beat.core.dataformat


#----------------------------------------------------------


class DataFormatCreationSerializer(ContributionCreationSerializer):
    class Meta(ContributionCreationSerializer.Meta):
        model = DataFormat
        beat_core_class = beat.core.dataformat


#----------------------------------------------------------


class BaseSerializer(DynamicFieldsSerializer):
    name = serializers.CharField(source="fullname")
    accessibility = serializers.SerializerMethodField()

    class Meta:
        default_fields = ['name', 'short_description', 'accessibility']

    def get_accessibility(self, obj):
        return obj.get_sharing_display().lower()


#----------------------------------------------------------


class ReferencedDataFormatSerializer(BaseSerializer):
    class Meta(BaseSerializer.Meta):
        model = DataFormat
        default_fields = BaseSerializer.Meta.default_fields + ['creation_date']


#----------------------------------------------------------


class ReferencingSerializer(BaseSerializer):
    is_owner = serializers.SerializerMethodField()

    class Meta(BaseSerializer.Meta):
        default_fields = BaseSerializer.Meta.default_fields + ['is_owner', 'version',]

    def get_is_owner(self, obj):
        if hasattr(obj, 'author'):
            return obj.author == self.context.get('user')


#----------------------------------------------------------


class ReferencingDataFormatSerializer(ReferencingSerializer):
    class Meta(ReferencingSerializer.Meta):
        model = DataFormat


#----------------------------------------------------------


class ReferencingAlgorithmSerializer(ReferencingSerializer):
    class Meta(ReferencingSerializer.Meta):
        model = Algorithm


#----------------------------------------------------------


class ReferencingDatabaseSerializer(VersionableSerializer):
    class Meta:
        model = Database
        default_fields = ['name', 'short_description']


#----------------------------------------------------------


class DataFormatSerializer(ContributionSerializer):
    extend = serializers.SerializerMethodField()
    referenced_dataformats = ReferencedDataFormatSerializer(many=True, source="referenced_formats")
    needed_dataformats = ReferencedDataFormatSerializer(many=True, source='all_needed_dataformats')
    referencing_dataformats = serializers.SerializerMethodField()
    referencing_algorithms = serializers.SerializerMethodField()
    referencing_databases = serializers.SerializerMethodField()

    class Meta(ContributionSerializer.Meta):
        model = DataFormat
        default_fields = ContributionSerializer.Meta.default_fields + ['extend']

    def get_extend(self, obj):
        return obj.extend.fullname() if obj.extend else None


    def get_referencing_dataformats(self, obj):
        user = self.context.get('user')
        dataformats = obj.referencing.for_user(user, True)
        serializer = ReferencingDataFormatSerializer(dataformats, many=True, context=self.context)
        return serializer.data


    def get_referencing_algorithms(self, obj):
        user = self.context.get('user')
        algorithms = Algorithm.objects.for_user(user, True).filter(endpoints=obj.algorithm_endpoints.all()).distinct()
        serializer = ReferencingAlgorithmSerializer(algorithms, many=True, context=self.context)
        return serializer.data


    def get_referencing_databases(self, obj):
        databaseset = DatabaseSet.objects.filter(template__outputs=obj.database_outputs.distinct())
        databases = Database.objects.filter(protocols__sets=databaseset).distinct().order_by('name')
        serializer = ReferencingDatabaseSerializer(databases, many=True, context=self.context)
        return serializer.data


#----------------------------------------------------------


class FullDataFormatSerializer(DataFormatSerializer):

    class Meta(DataFormatSerializer.Meta):
        default_fields = DataFormatSerializer.Meta.default_fields + DataFormatSerializer.Meta.extra_fields
