#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.http import HttpResponseRedirect
from django.core.urlresolvers import reverse
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.views.decorators.csrf import csrf_protect
from django.contrib.auth.decorators import login_required
from django.conf import settings
from django.contrib.auth.models import User

from .forms import CreationForm, UpdateForm
from .models import DataFormat
from ..team.models import Team


#----------------------------------------------------------


DEFAULT_DATAFORMAT_TEXT = """{
}"""


#----------------------------------------------------------


@csrf_protect
@login_required
def create(request, name=None):
    """Creates a new data format or a new version of an existing data format

    The user must be authenticated before it can add a new data format
    """

    # Process the form
    form = None
    form_args = dict(
                    request_user=request.user,
                    name=name,
                )


    # Retrieves the existing object if possible
    previous_version = None
    version = 1
    if name is not None:
        try:
            previous_version = DataFormat.objects.filter(
                                    author=request.user,
                                    name__iexact=name,
                                ).order_by('-version')[0]
            version = previous_version.version + 1
        except:
            pass


    if request.method == 'POST':
        form = CreationForm(request.POST, **form_args)

        if form.is_valid():
            (dataformat, errors) = DataFormat.objects.create_dataformat(
                author=request.user,
                name=form.cleaned_data['name'],
                version=form.cleaned_data['version'],
                short_description=form.cleaned_data['short_description'],
                description=form.cleaned_data['description'],
                declaration=form.cleaned_data['file'],
                previous_version=previous_version
            )

            if dataformat is not None:
                # Redirect to the page of the dataformat just created
                return HttpResponseRedirect(
                        reverse('dataformats:view',
                            args=(
                                dataformat.author.username,
                                dataformat.name,
                                dataformat.version,
                            ),
                        )
                    )

            for error in errors.split('\n'):
                form.add_error('file', error)


    if form is None:
        if previous_version is not None:
            # fill-in with previous version's settings
            form_args['initial'] = {
                'name': previous_version.name,
                'version': version,
                'short_description': previous_version.short_description,
                'description': previous_version.description,
                'file': previous_version.declaration_string
            }
        else:
            # start from scratch
            form_args['initial'] = {
                'version': 1,
                'file': DEFAULT_DATAFORMAT_TEXT,
            }
            if form_args['name']:
                form_args['initial']['name'] = form_args['name']

        form = CreationForm(**form_args)


    return render_to_response('dataformats/edition.html',
                              {
                                'form': form,
                                'object': previous_version,
                                'op': 'new-version' if previous_version else 'new',
                                'edition': False,
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


@csrf_protect
@login_required
def fork(request, author, name, version):
    """Creates a new data format by forking an existing data format

    The user must be authenticated before it can add a new data format
    """

    # Process the form
    form = None
    form_args = dict(
                    request_user=request.user,
                )


    # Retrieves the forked data format
    fork_of = get_object_or_404(DataFormat.objects.for_user(request.user, True),
                                author__username__iexact=author,
                                name__iexact=name,
                                version=int(version)
                               )

    # Handle the posting of the form
    if request.method == 'POST':
        form = CreationForm(request.POST, **form_args)

        if form.is_valid():
            (dataformat, errors) = DataFormat.objects.create_dataformat(
                author=request.user,
                name=form.cleaned_data['name'],
                short_description=form.cleaned_data['short_description'],
                description=form.cleaned_data['description'],
                declaration=form.cleaned_data['file'],
                fork_of=fork_of
            )

            if dataformat is not None:
                # Redirect to the page of the dataformat just created
                return HttpResponseRedirect(
                        reverse('dataformats:view',
                            args=(
                                dataformat.author.username,
                                dataformat.name,
                                dataformat.version,
                            ),
                        )
                    )

            for error in errors.split('\n'):
                form.add_error('file', error)


    # Initialise the form if necessary
    if form is None:
        form_args['initial'] = {
            'version': 1,
            'short_description': fork_of.short_description,
            'description': fork_of.description,
            'file': fork_of.declaration_string,
        }

        form = CreationForm(**form_args)


    return render_to_response('dataformats/edition.html',
                              {
                                'form': form,
                                'object': fork_of,
                                'op': 'fork',
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


@csrf_protect
@login_required
def edit(request, author, name, version):
    """Allows the modification of the data format

    The user must be authenticated before it can edit a data format
    """

    if author != request.user.username:
        raise Http404()

    # Process the form
    form = None
    form_args = dict()


    # Retrieves the forked data format
    dataformat = get_object_or_404(DataFormat,
                                   author__username__iexact=author,
                                   name__iexact=name,
                                   version=int(version)
                                  )


    # Handle the posting of the form
    if request.method == 'POST':
        form = UpdateForm(request.POST, **form_args)

        if form.is_valid():
            dataformat.declaration = form.cleaned_data['file']
            dataformat.description = form.cleaned_data['description']

            declaration = dataformat.declaration
            declaration['#description'] = form.cleaned_data['short_description']
            dataformat.declaration = declaration

            try:
                dataformat.save()
            except SyntaxError as e:
                for arg in e.args:
                    form.add_error('file', arg)
            else:
                # Redirect to the page of the dataformat just created
                return HttpResponseRedirect(
                        reverse('dataformats:view',
                            args=(
                                dataformat.author.username,
                                dataformat.name,
                                dataformat.version,
                            ),
                        )
                    )


    # Initialise the form if necessary
    if form is None:
        form_args['initial'] = {
            'version': 1,
            'short_description': dataformat.short_description,
            'description': dataformat.description,
            'file': dataformat.declaration_string,
        }

        form = UpdateForm(**form_args)

    return render_to_response('dataformats/edition.html',
                              {'form': form,
                               'op': 'edit',
                               'object': dataformat,
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def view(request, author, name, version=None):
    """Shows the dataformat"""

    # Retrieves the dataformat
    if version:
        dataformat = get_object_or_404(
                DataFormat,
                author__username__iexact=author,
                name__iexact=name,
                version=int(version),
                )
    else:
        dataformat = DataFormat.objects.filter(
                author__username__iexact=author,
                name__iexact=name).order_by('-version')
        if not dataformat:
            raise Http404()
        else:
            dataformat = dataformat[0]

    (has_access, _) = dataformat.accessibility_for(request.user)

    if not has_access: raise Http404()

    owner = (request.user == dataformat.author)

    # Users the object can be shared with
    users = User.objects.exclude(username__in=settings.ACCOUNTS_TO_EXCLUDE_FROM_TEAMS).order_by('username')

    # Render the page
    return render_to_response('dataformats/view.html',
                              {
                                  'dataformat': dataformat,
                                  'owner': owner,
                                  'users': users,
                                  'teams': Team.objects.for_user(request.user, True)
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def diff(request, author1, name1, version1, author2, name2, version2):
    """Shows the difference between two data formats. The Web API is used to retrieve the
    details about the data formats and check the accessibility.

    Answers to the following URLs

    <dataformats>/diff/<author1>/<name1>/<version1>/<author2>/<name2>/<version2>/

    """

    return render_to_response('dataformats/diff.html',
                              {'dataformat1_author':  author1,
                               'dataformat1_name':    name1,
                               'dataformat1_version': version1,
                               'dataformat2_author':  author2,
                               'dataformat2_name':    name2,
                               'dataformat2_version': version2,
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def ls(request, author_name):
    '''List all accessible dataformats to the request user'''

    if not author_name: return public_ls(request)

    # check that the user exists on the system
    author = get_object_or_404(User, username=author_name)

    # orders dataformats so that the latest information is displayed first
    objects = DataFormat.objects.from_author_and_public(request.user,
            author_name).order_by('-creation_date')
    objects = DataFormat.filter_latest_versions(objects)

    return render_to_response('dataformats/list.html',
            dict(
                objects=objects,
                author=author,
                owner=(request.user==author),
                ),
            context_instance=RequestContext(request),
            )


#----------------------------------------------------------


def public_ls(request):
    '''List all publicly accessible data formats'''

    # orders so that latest is displayed first
    objects = DataFormat.objects.public().order_by('-creation_date')
    objects = DataFormat.filter_latest_versions(objects)

    return render_to_response('dataformats/list.html',
            dict(
                objects=objects,
                author=request.user, #anonymous
                owner=False,
                ),
            context_instance=RequestContext(request),
            )
