#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.db import transaction
from django.contrib.auth.models import User
from django.core.urlresolvers import reverse
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.core.mail import send_mail
from django.template.loader import render_to_string
from django.contrib.sites.models import Site

import beat.core.hash
import beat.core.data
import beat.core.algorithm
import beat.core.experiment

from beat.core.utils import NumpyJSONEncoder

from ..algorithms.models import Algorithm
from ..toolchains.models import Toolchain

from ..common.models import Shareable
from ..common.models import ContributionManager
from ..common.models import get_contribution_declaration_filename
from ..common.models import get_contribution_description_filename
from ..common.models import get_description
from ..common.models import set_description
from ..common import storage

from ..common.exceptions import ShareError
from ..common.texts import Messages
from ..common.storage import OverwriteStorage
from ..backend.models import Queue, Environment
from ..databases.models import DatabaseSet,  DatabaseSetOutput
from ..import __version__


from datetime import datetime

import os
import glob
import simplejson

import logging
logger = logging.getLogger(__name__)


#----------------------------------------------------------


def validate_environments(experiment):
    """Validates the environments throughout the experiment"""

    def _valid(queue, environment):
        return bool(Queue.objects.filter(name=queue,
            environments__name=environment['name'],
            environments__version=environment['version']))

    errors = []

    default_q = experiment.data['globals']['queue']
    default_env = experiment.data['globals']['environment']
    if not _valid(default_q, default_env):
        errors.append("The combination of queue '%s' with environment '%s (%s)' in the global experiment declaration does not exist" % (default_q, default_env['name'], default_env['version']))

    for name, config in experiment.blocks.items():
        q = config.get('queue', default_q)
        env = config.get('environment', default_env)
        if not _valid(q, env):
            errors.append("The combination of queue '%s' with environment '%s (%s)' for block '%s' does not exist" % (q, env['name'], env['version'], name))

    for name, config in experiment.analyzers.items():
        q = config.get('queue', default_q)
        env = config.get('environment', default_env)
        if not _valid(q, env):
            errors.append("The combination of queue '%s' with environment '%s (%s)' for analyzer '%s' does not exist" % (q, env['name'], env['version'], name))

    return errors


def validate_experiment(experiment_info, toolchain_info):
    """Makes sure the experiment can be run"""

    xp = beat.core.experiment.Experiment(settings.PREFIX,
            (experiment_info, toolchain_info))

    if not xp.valid: return xp, xp.errors

    return xp, xp.errors + validate_environments(xp)


#----------------------------------------------------------


class DeclarationStorage(OverwriteStorage):

    def __init__(self, *args, **kwargs):
        super(DeclarationStorage, self).__init__(*args, location=settings.EXPERIMENTS_ROOT, **kwargs)


# Keep for backward compatibility with initial migration
class ConfigurationStorage(OverwriteStorage):

    def __init__(self, *args, **kwargs):
        super(ConfigurationStorage, self).__init__(*args, location=settings.EXPERIMENTS_ROOT, **kwargs)

#----------------------------------------------------------


class ExperimentManager(ContributionManager):

    def get_by_natural_key(self, author_username, toolchain_username,
            toolchain_name, toolchain_version, name):
        return self.get(
                author__username=author_username,
                toolchain__author__username=toolchain_username,
                toolchain__name=toolchain_name,
                toolchain__version=toolchain_version,
                name=name
                )

    def from_author(self, user, author_name, add_public=False):
        return super(ExperimentManager, self).from_author(user, author_name, add_public).order_by('-creation_date', 'name')

    def from_author_and_public(self, user, author_name):
        return super(ExperimentManager, self).from_author_and_public(user, author_name).order_by('-creation_date', 'name')

    def create_experiment(self, author, toolchain, name, declaration,
                          short_description='', description=''):
        """Creates a new experiment in pending state"""

        # Create the database representation of the experiment
        experiment = self.model(
            author = author,
            toolchain = toolchain,
            name = name,
            status = self.model.PENDING,
            start_date = None,
            end_date = None,
        )

        if not(isinstance(declaration, dict)):
            declaration = simplejson.loads(declaration)

        if len(short_description) > 0:
            declaration['description'] = short_description

        experiment.declaration = declaration

        # Check the provided description
        if description is not None:
            experiment.description = description

        # Save the experiment (will run the validation)
        try:
            experiment.save()
        except Exception:
            import traceback
            return (None, None, traceback.format_exc())

        experiment._loaded_status = experiment.status
        return experiment, experiment._toolchain_cache, None


#----------------------------------------------------------


def get_contribution_configuration_filename(obj, path):
    return obj.configuration_filename()


#----------------------------------------------------------


class Experiment(Shareable):

    #_____ Constants __________

    PENDING   = 'P'
    SCHEDULED = 'S'
    RUNNING   = 'R'
    DONE      = 'D'
    FAILED    = 'F'
    CANCELING = 'C'

    STATUS = (
            (PENDING,   'Pending'),
            (SCHEDULED, 'Scheduled'),
            (RUNNING,   'Running'),
            (DONE,      'Done'),
            (FAILED,    'Failed'),
            (CANCELING, 'Canceling'),
    )


    #_____ Fields __________

    author            = models.ForeignKey(User, related_name='experiments',
        on_delete=models.CASCADE)
    toolchain         = models.ForeignKey(Toolchain,
        related_name='experiments', on_delete=models.CASCADE)
    name              = models.CharField(max_length=200)
    short_description = models.CharField(max_length=100, default='', blank=True, help_text=Messages['short_description'])
    status            = models.CharField(max_length=1, choices=STATUS, default=PENDING)
    creation_date     = models.DateTimeField(null=True, blank=True, auto_now_add=True)
    start_date        = models.DateTimeField(null=True, blank=True)
    end_date          = models.DateTimeField(null=True, blank=True)

    declaration_file = models.FileField(
                                storage=DeclarationStorage(),
                                upload_to=get_contribution_declaration_filename,
                                blank=True, null=True,
                                max_length=300,
                                db_column='declaration'
                               )

    description_file = models.FileField(
                                storage=DeclarationStorage(),
                                upload_to=get_contribution_description_filename,
                                blank=True, null=True,
                                max_length=300,
                                db_column='description'
                               )


    # read-only parameters that are updated at every save(), if required
    hash = models.CharField(max_length=64)
    referenced_datasets = models.ManyToManyField(DatabaseSet, related_name='experiments', blank=True)
    referenced_algorithms = models.ManyToManyField(Algorithm, related_name='experiments', blank=True)

    objects = ExperimentManager()


    #_____ Meta parameters __________

    class Meta:
        unique_together = ('author', 'toolchain', 'name')
        ordering = ['-creation_date']


    #_____ Utilities __________

    def natural_key(self):

        return (self.author.username,) + \
                self.toolchain.natural_key() + (self.name,)


    #_____ Methods __________

    def fullname(self):
        return '%s/%s/%s' % (self.author.username, self.toolchain.fullname(), self.name)


    def hashed_path(self, extension=''):
        """Relative path of a file belonging to the object on the respective
        storage"""

        return os.path.join(
                beat.core.hash.toUserPath(self.author.username),
                self.toolchain.fullname(),
                self.name + extension,
               )


    def declaration_filename(self):
        """Relative path of the declaration file on the storage"""
        return self.hashed_path('.json')


    def description_filename(self):
        """Relative path of the description file on the storage"""
        return self.hashed_path('.rst')


    #_____ Utilities __________

    def _share_dataformats(self, users, teams):
        needed_formats = []
        for dataset in self.referenced_datasets.all():
            needed_formats.extend(dataset.all_needed_dataformats())

        # Retrieve and process the list of referenced dataformats
        own_needed_formats = filter(lambda x: x.author == self.author, needed_formats)
        other_needed_formats = filter(lambda x: x.author != self.author, needed_formats)

        # Ensure that all needed dataformats from other users/teams have the necessary sharing
        # preferences
        errors = []
        for needed_format in other_needed_formats:
            errors.extend(needed_format.is_accessible(users=users, teams=teams))

        if len(errors) > 0:
            raise ShareError(errors)

        for dataformats in own_needed_formats:
            dataformats.share(users=users, teams=teams)


    def __share_algorithms(self, users, teams, algorithms_infos):
        # Retrieve and process the list of algorithms referenced by the declaration, and
        # the data formats they reference
        needed_algorithms = []
        for algorithm in self.referenced_algorithms.iterator():
            needed_algorithms.append(algorithm)
        needed_algorithms = list(set(needed_algorithms))

        # Only keep the referenced dataformats and algorithms owned by the author of the
        # experiment
        # Process the list of referenced algorithms
        own_needed_algorithms = filter(lambda x: x.author == self.author, needed_algorithms)
        other_needed_algorithms = filter(lambda x: x.author != self.author, needed_algorithms)

        # Ensure that all needed algorithms from other users have the necessary sharing
        # preferences
        errors = []
        for needed_algorithm in other_needed_algorithms:
            errors.extend(needed_algorithm.is_accessible(False, users, teams))
        if len(errors) > 0:
            raise ShareError(errors)

        for algorithm in own_needed_algorithms:
            if algorithms_infos and algorithms_infos.has_key(algorithm.fullname()):
                infos = algorithms_infos[algorithm.fullname()]
                opensource = infos.get('opensource', False)
            else:
                opensource = True

            algorithm.share(public=opensource, users=users, teams=teams)

    #_____ Overrides __________

    def has_attestation(self):
        try:
            self.attestation
        except ObjectDoesNotExist:
            return False
        else:
            return True

    @classmethod
    def from_db(cls, db, field_names, values):
        instance = super(Experiment, cls).from_db(db, field_names, values)
        instance._loaded_status = values[field_names.index('status')]
        return instance

    def save(self, *args, **kwargs):
        new_experiment = (self.id is None)

        # if changing the experiment status, then make sure start_date is set
        if (self.status not in [self.PENDING, self.SCHEDULED]) and \
                (self.start_date is None):
            self.start_date = self.end_date or datetime.now()

        # Retrieve the experiment declaration
        declaration = self.declaration

        # Compute the hash of the content
        content_hash = beat.core.hash.hashJSON(declaration, 'description')
        content_modified = (content_hash != self.hash)

        if content_modified:
            # validates the experiment
            xp, errors = validate_experiment(declaration, self.toolchain.declaration)
            if errors:
                message = "The experiment isn't valid, due to the " \
                        "following errors:\n  * %s"
                raise SyntaxError(message % '\n  * '.join(errors))

            self.hash = content_hash

        self.short_description = declaration.get('description', '')

        # Ensures that the sharing informations are consistent
        if self.sharing == Shareable.USABLE:
            self.sharing = Shareable.PUBLIC

        # Save the changed files (if necessary)
        storage.save_files(self)

        is_adding = self._state.adding

        if not is_adding and self._loaded_status != self.status:
            if self.status in [Experiment.DONE, Experiment.FAILED]:
                self.email()

        # Invoke the base implementation
        super(Experiment, self).save(*args, **kwargs)

        # If the filename has changed, move all the files
        if self.declaration_filename() != self.declaration_file.name:
            storage.rename_file(self, 'declaration_file', self.declaration_filename())
            storage.rename_file(self, 'description_file', self.description_filename())

        if content_modified:
            # Creates experiment blocks and setup dependencies
            self.update_blocks()

            # Link the experiment to the datasets
            self.referenced_datasets.clear()
            for dataset_declaration in xp.datasets.values():
                try:
                    (db_name, db_version) = \
                            dataset_declaration['database'].name.split('/')
                    dataset = DatabaseSet.objects.get(
                            protocol__database__name=db_name,
                            protocol__database__version=db_version,
                            protocol__name=dataset_declaration['protocol'],
                            name=dataset_declaration['set'],
                            )
                except:
                    if new_experiment: self.delete()
                    raise SyntaxError("The dataset '%s.%s.%s' can't be found" \
                            % (
                                dataset_declaration['database'].name,
                                dataset_declaration['protocol'],
                                dataset_declaration['set'],
                                )
                            )

                self.referenced_datasets.add(dataset)

            # Link the experiment to the algorithms
            self.referenced_algorithms.clear()

            for algorithm in xp.algorithms:
                try:
                    (username, name, version) = algorithm.split('/')
                    algorithm_db = Algorithm.objects.get(
                            author__username=username,
                            name=name,
                            version=int(version),
                            )
                except:
                    if new_experiment: self.delete()
                    raise SyntaxError("The algorithm '%s' can't be found" % \
                            algorithm.fullname())

                self.referenced_algorithms.add(algorithm_db)


    def email(self):
        '''e-mails owners and shared parties about this experiment status'''

        user_email_list = [self.author.email] if self.author.accountsettings.experiment_mail_notifications_enabled else []
        user_email_list.extend([user.email for user in self.shared_with.all() if user.accountsettings.experiment_mail_notifications_enabled])
        all_team_members = [user for team in self.shared_with_team.all() for user in team.members.all()]
        user_email_list.extend([user.email for user in all_team_members if user.email not in user_email_list and user.accountsettings.experiment_mail_notifications_enabled])

        if user_email_list:
            if self.status == Experiment.DONE:
                subject = "Experiment %s finished successfully" % \
                    self.fullname()
                template_path = 'experiments/successful_experiment_email.txt'

            elif self.status == Experiment.FAILED:
                subject = "Experiment %s failed" % self.fullname()
                template_path = 'experiments/failed_experiment_email.txt'

            try:
                send_mail(subject, render_to_string(template_path, {'experiment': self, 'beat_version': __version__, 'site': Site.objects.get_current()}), settings.DEFAULT_FROM_EMAIL, user_email_list)
            except Exception:
                import traceback
                logger.warn("Could not send e-mail to `%s' about " \
                    "`%s'. Exception caught: %s", user_email_list,
                    self, traceback.format_exc())


    def share(self, users=None, teams=None, algorithms_infos={}):
        self._share_dataformats(users=users, teams=teams)
        self.__share_algorithms(users=users, teams=teams, algorithms_infos=algorithms_infos)
        self.toolchain.share(users=users, teams=teams)
        super(Experiment, self).share(users=users, teams=teams)


    def update_blocks(self):
        """Updates internal block representation of an experiment"""

        corexp = self.core()

        # Loads the experiment execution description, creating the Block's,
        # BlockInput's and BlockOutput's as required.
        for order_0, (block_name, description) in \
            enumerate(corexp.setup().items()):

            # Checks that the Queue/Environment exists
            job_description = description['configuration']

            env = Environment.objects.filter(
                name=job_description['environment']['name'],
                version=job_description['environment']['version'],
                )

            if not env:
                logger.warn("Cannot find environment `%s (%s)' - not setting",
                    job_description['environment']['name'],
                    job_description['environment']['version'])
                env = None
            else:
                env = env[0]

            # Search for queue that contains a specific environment
            if env:
              queue = Queue.objects.filter(name=job_description['queue'],
                environments__in=[env])
            else:
              queue = Queue.objects.filter(name=queue)
            if not queue:
                env_name = env.fullname() if env else 'NULL'
                logger.warn("Cannot find queue `%s' which contains " \
                    "environment `%s' - not setting",
                    job_description['queue'], env_name)
                queue = None
            else:
                queue = queue[0]

            parts = job_description['algorithm'].split('/')
            algorithm = Algorithm.objects.get(
                author__username=parts[0],
                name=parts[1],
                version=parts[2],
                )

            # Ties the block in
            slots = job_description.get('nb_slots')
            b = Block.objects.filter(experiment=self, name=block_name).first()
            if b is None:
                b = Block(experiment=self, name=block_name, algorithm=algorithm)
            else:
                b.algorithm = algorithm
            b.execution_order = order_0 + 1
            b.command=simplejson.dumps(job_description, indent=4)
            b.status=Block.NOT_CACHED
            b.analyzer=algorithm.analysis()
            b.environment=env
            b.queue=queue
            b.required_slots=job_description['nb_slots']
            b.channel=job_description['channel']
            b.save()

            # from this point: requires block to have an assigned id
            b.dependencies.clear()
            b.dependencies.add(*[self.blocks.get(name=k) \
                for k in description['dependencies']])

            # reset inputs and outputs - creates if necessary only
            b.inputs.clear()
            for v in job_description['inputs'].values():
                if 'database' in v: #database input
                    db = DatabaseSetOutput.objects.get(hash=v['hash'])
                    BlockInput.objects.get_or_create(block=b,
                        channel=v['channel'], database=db)
                else:
                    cache = CachedFile.objects.get(hash=v['hash'])
                    BlockInput.objects.get_or_create(block=b,
                        channel=v['channel'], cache=cache)

            b.outputs.clear()
            outputs = job_description.get('outputs',
                {'': job_description.get('result')})
            for v in outputs.values():
                cache, cr = CachedFile.objects.get_or_create(hash=v['hash'])
                cache.blocks.add(b)


    #_____ Methods __________

    def is_busy(self):
        return self.status in [Experiment.PENDING, Experiment.SCHEDULED, Experiment.CANCELING]

    def is_done(self):
        return self.status in [Experiment.DONE, Experiment.FAILED]

    def is_running(self):
        return self.status == Experiment.RUNNING

    def modifiable(self):
        return (self.reports.count() == 0) and not self.has_attestation() and super(Experiment, self).modifiable()

    def deletable(self):
        return (self.reports.count() == 0) and not self.has_attestation() and super(Experiment, self).deletable()

    def core(self):
        return validate_experiment(self.declaration, self.toolchain.declaration)[0]

    def job_splits(self, status=None):
        from ..backend.models import JobSplit
        retval = JobSplit.objects.filter(job__block__in=self.blocks.all())
        if status is not None:
            retval = retval.filter(status=status)
        return retval

    def get_absolute_url(self):
        return reverse(
                'experiments:view',
                args=(
                    self.author.username,
                    self.toolchain.author.username,
                    self.toolchain.name,
                    self.toolchain.version,
                    self.name,
                    ),
                )

    def get_api_share_url(self):
        return reverse(
                'api_experiments:share',
                args=(
                    self.author.username,
                    self.toolchain.author.username,
                    self.toolchain.name,
                    self.toolchain.version,
                    self.name,
                    ),
                )

    def get_api_update_url(self):
        return reverse(
                'api_experiments:object',
                args=(
                    self.author.username,
                    self.toolchain.author.username,
                    self.toolchain.name,
                    self.toolchain.version,
                    self.name,
                    ),
                )

    def get_admin_change_url(self):
        return reverse('admin:experiments_experiment_change', args=(self.id,))

    def completion(self):
        if self.start_date is None:
            return 0

        if self.end_date is not None:
            return 100

        blocks = self.blocks.all()

        if len(blocks) == 0:
            return 0

        return int(100 * float(len(filter(lambda x: x.status == Block.CACHED, blocks))) / len(blocks))

    def all_needed_dataformats(self):
        result = []

        for algorithm in self.referenced_algorithms.all():
            result.extend(algorithm.all_needed_dataformats())

        for dataset in self.referenced_datasets.all():
            result.extend(dataset.all_needed_dataformats())

        return list(set(result))


    def reset(self):
        """Resets an experiment so it can be run again"""

        if not self.is_done(): return #can only reset experiments which are done

        self.blocks.update(
            status=Block.NOT_CACHED,
            start_date=None,
            end_date=None,
            )

        self.start_date = None
        self.end_date = None
        self.status = self.PENDING

        # reset sharing state
        self.sharing    = Shareable.PRIVATE
        self.shared_with.clear()
        self.shared_with_team.clear()

        # remove associated attestations
        if self.has_attestation(): self.attestation.all().delete()

        self.save()


    def databases_and_protocols(self):
        '''A set of all used database/protocol combinations for all datasets'''

        return set(['%s' % k.protocol for k in self.referenced_datasets.all()])


    def analyzers(self):
        '''A list of all used analyzers'''

        return set([k.fullname() for k in self.referenced_algorithms.all() if k.analysis()])


    #_____ Properties __________

    def set_declaration(self, value):
        if isinstance(value, dict):
            value = simplejson.dumps(value,
                                     indent=4,
                                     cls=NumpyJSONEncoder)

        storage.set_file_content(self, 'declaration_file', self.declaration_filename(), value)

    def get_declaration(self):
        return simplejson.loads(storage.get_file_content(self, 'declaration_file'))

    declaration = property(get_declaration, set_declaration)

    description = property(get_description, set_description)


    @property
    def declaration_string(self):
        return storage.get_file_content(self, 'declaration_file')


    def _update_state(self):
        '''Update self state based on associated block states

        This method is called by the underlying block. It is not part of the
        Experiment's public API and must not be called by any other user code.
        '''

        self_ = Experiment.objects.select_for_update().get(pk=self.pk)

        if self_.is_done(): return

        if self.start_date is None:
            d = self.blocks.filter(start_date__isnull=False).\
                order_by('start_date')
            if d:
                self.start_date = d.first().start_date
            else:
                self.start_date = datetime.now()

        block_statuses = self.blocks.values_list('status', flat=True)

        # Process main state and state from job results
        if Block.FAILED in block_statuses or Block.CANCELLED in block_statuses:
            if Block.PROCESSING in block_statuses:
                self.status = Experiment.CANCELING
            else:
                self.status = Experiment.FAILED

        elif (Block.PROCESSING in block_statuses) or \
            ((Block.NOT_CACHED in block_statuses or \
            Block.SKIPPED in block_statuses) and \
            Block.CACHED in block_statuses):
            self.status = Experiment.RUNNING

        elif Block.NOT_CACHED not in block_statuses:
            self.status = Experiment.DONE

        else:
            self.status = Experiment.SCHEDULED

        # Set end date if experiment is done
        if self.is_done() and self.end_date is None:
            d = self.blocks.filter(end_date__isnull=False).\
                order_by('-end_date')
            if d:
                self.end_date = d.first().end_date
            else:
                self.end_date = datetime.now()


        self.save()


    @transaction.atomic
    def schedule(self):
        '''Schedules this experiment for execution at the backend

        Because the experiment is fully built on ``save()`` (including block
        interdependence and cache requirements), to "schedule" means solely
        creating :py:class:`..backend.models.Job`'s to address all
        algorithm-equipped blocks in the experiment. A ``Job`` is the
        reflection of the experiment's block for the backend and makes the
        schedule aware of execution units that must be processed. Each ``Job``
        is then split on the scheduler process, for as many times as required
        by the :py:class:`Block`'s ``required_slots`` entry, effectively
        creating one :py:class:`..backend.models.JobSplit` per split.
        '''

        self_ = Experiment.objects.select_for_update().get(pk=self.pk)

        if self_.status != Experiment.PENDING: return

        for b in self.blocks.all(): b._schedule()

        # notice that the previous call may decide all is done already
        # so, we must respect that before setting the SCHEDULED status
        self.refresh_from_db()
        if not self.is_done():
            self.status = Experiment.SCHEDULED
            self.save()


    @transaction.atomic
    def cancel(self):
        '''Cancels the execution of this experiment on the backend.

        .. caution::

           After each block is scheduled, it is possible some or all splits for
           a given block are under execution. We must select-for-update all
           Blocks and associated Jobs, as to avoid concurrent resetting from a
           separate scheduling process.

        '''

        self_ = Experiment.objects.get(pk=self.pk)

        if self_.status not in (Experiment.SCHEDULED, Experiment.RUNNING):
            return

        with transaction.atomic():
            for b in self.blocks.all(): b._cancel()


    def fork(self, username=None, name=None):
        '''Forks this experiment under a new username or name'''

        author = username or self.author
        name = name or self.name
        xp, _, __ = Experiment.objects.create_experiment(author,
            self.toolchain, name, self.get_declaration(),
            self.short_description, self.description)
        return xp


#----------------------------------------------------------


class BlockManager(models.Manager):

    def get_by_natural_key(self, name, experiment_author,
        toolchain_author, toolchain_name,
        toolchain_version, experiment_name):
        return self.get(
            name=name,
            experiment__author__username=experiment_author,
            experiment__toolchain__author__username=toolchain_author,
            experiment__toolchain__name=toolchain_name,
            experiment__toolchain__version=toolchain_version,
            experiment__name=experiment_name,
            )


class Block(models.Model):

    NOT_CACHED = 'N'
    PROCESSING = 'P'
    CACHED     = 'C'
    FAILED     = 'F'
    SKIPPED    = 'S'
    CANCELLED  = 'L'

    STATUS = (
        (NOT_CACHED, 'Not cached'),
        (PROCESSING, 'Processing'),
        (CACHED,     'Cached'),
        (FAILED,     'Failed'),
        (SKIPPED,    'Skipped'),
        (CANCELLED,  'Cancelled'),
    )

    experiment = models.ForeignKey(Experiment, related_name='blocks',
        on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    command = models.TextField(null=True, blank=True)
    status = models.CharField(max_length=1, choices=STATUS, default=NOT_CACHED)
    analyzer = models.BooleanField(default=False)
    algorithm = models.ForeignKey(Algorithm, related_name='blocks',
        on_delete=models.CASCADE)
    creation_date = models.DateTimeField(null=True, blank=True,
        auto_now_add=True)
    start_date = models.DateTimeField(null=True, blank=True)
    end_date = models.DateTimeField(null=True, blank=True)
    environment = models.ForeignKey(Environment, related_name='blocks',
        null=True, on_delete=models.SET_NULL)
    queue = models.ForeignKey(Queue, related_name='blocks', null=True,
        on_delete=models.SET_NULL)

    required_slots = models.PositiveIntegerField(default=1)
    channel = models.CharField(max_length=200, default='', blank=True,
        help_text="Synchronization channel within the toolchain")

    # relationship to blocks to which this block depends on
    dependencies = models.ManyToManyField('self',
                                          related_name='dependents',
                                          blank=True,
                                          symmetrical=False,
                                         )

    # order of this block within the experiment - useful for the `backup'
    # command, so we can dump the blocks in the right dependence order
    execution_order = models.PositiveIntegerField(null=True, blank=True)

    objects = BlockManager()


    class Meta:
        unique_together = ('experiment', 'name')

        # setup ordering so that the dump order respects self dependencies
        ordering = ['experiment_id', 'execution_order']


    def __str__(self):
        return self.experiment.fullname() + ', ' + self.name + ' (%s)' % self.get_status_display()

    def natural_key(self):
        return (
            self.name,
            self.experiment.author.username,
            self.experiment.toolchain.author.username,
            self.experiment.toolchain.name,
            self.experiment.toolchain.version,
            self.experiment.name,
            )

    # Accessors for statistics

    def __return_first__(self, field, default=None):
        return getattr(self.outputs.first(), field, default)

    def first_cache(self):
        return self.outputs.first()

    def error_report(self):
        return self.__return_first__('error_report')

    def stdout(self):
        return self.__return_first__('stdout')

    def stderr(self):
        return self.__return_first__('stderr')

    def speed_up_real(self):
        return self.__return_first__('speed_up_real')

    def speed_up_maximal(self):
        return self.__return_first__('speed_up_maximal')

    def linear_execution_time(self):
        return self.__return_first__('linear_execution_time')

    def queuing_time(self):
        return self.__return_first__('queuing_time')

    def cpu_time(self):
        return self.__return_first__('cpu_time')

    def max_memory(self):
        return self.__return_first__('max_memory')

    def data_read_size(self):
        return self.__return_first__('data_read_size')

    def data_read_nb_blocks(self):
        return self.__return_first__('data_read_nb_blocks')

    def data_read_time(self):
        return self.__return_first__('data_read_time')

    def data_written_size(self):
        return self.__return_first__('data_written_size')

    def data_written_nb_blocks(self):
        return self.__return_first__('data_written_nb_blocks')

    def data_written_time(self):
        return self.__return_first__('data_written_time')

    # Accessor for results
    results = property(lambda self: self.__return_first__('results'))


    def _schedule(self):
        '''Schedules this block for execution at the backend

        To "schedule" means solely creating a :py:class:`..backend.models.Job`
        pointing to this object. This method **should only be called by the
        owning experiment**. It is not part of the Block's public API.
        '''

        # lock self - avoids concurrent update from scheduler/worker subsystem
        self_ = Block.objects.select_for_update().get(pk=self.pk)

        # checks we have not, meanwhile, been cancelled
        if self_.done(): return

        # checks queue and environment
        if self.queue is None:
            raise RuntimeError("Block `%s' does not have a queue assigned " \
                "- this normally indicates the originally selected " \
                "queue was deleted since the experiment was first " \
                "configured. Re-configure this experiment and select a new " \
                "default or block-specific queue" % self.name)

        if self.environment is None:
            raise RuntimeError("Block `%s' does not have an environment " \
                "assigned - this normally indicates the originally selected " \
                "environment was deleted since the experiment was first " \
                "configured. Re-configure this experiment and select a new " \
                "default or block-specific environment" % self.name)

        from ..backend.models import Job

        # search for other jobs with similar outputs that have no children yet
        # do this carefully, as other experiments may be scheduled at the same
        # time, invalidating our "parent" choice
        parent = Job.objects.filter(block__outputs__in=self.outputs.all(),
            child=None).first()
        if parent is not None: #(candidate only) try to lock it
            while True:
                parent = Job.objects.select_for_update().get(pk=parent.pk)
                if parent.child_ is not None: #was taken meanwhile, retry
                    parent = parent.child
                    continue
                Job(block=self, parent=parent).save()
                break
        else:
            Job(block=self).save()

        # checks if the job is immediately runnable - if so, tries to
        # make it runnable (check caches and other)
        if self.is_runnable(): self.job._make_runnable()


    def done(self):
      '''Says whether the block has finished or not'''

      return self.status not in (Block.NOT_CACHED, Block.PROCESSING)


    def _cancel(self):
        '''Cancels the execution of this block on the backend.

        This method should only be called from the experiment equivalent. It is
        not part of the Block's public API.
        '''

        # lock self - avoids concurrent update from scheduler/worker subsystem
        self_ = Block.objects.select_for_update().get(pk=self.pk)

        if self_.done(): return

        if hasattr(self, 'job'):
            self.job._cancel()
        else:
            self.status = Block.CANCELLED
            self.save()
            self.experiment._update_state()


    def is_runnable(self):
        '''Checks if a block is runnable presently'''

        return all([k.status in (Block.CACHED, Block.SKIPPED) \
                for k in self.dependencies.all()]) and \
                (hasattr(self, 'job') and self.job.parent is None)


    def _cascade_updates(self):
        '''Cascade updates to blocks once I'm done.
        '''

        for b in self.dependents.all():
            if any([k.status in (Block.FAILED, Block.CANCELLED) \
                for k in b.dependencies.all()]):
                    b._cancel()
            if b.is_runnable(): b.job._make_runnable()

        # Update eventual running siblings in case of a failure
        if self.status == Block.FAILED:
            for b in Block.objects.filter(experiment=self.experiment,
                status=Block.PROCESSING):
                b._cancel()


    def _update_state(self, timings=None):
        '''Updates self state as a result of backend running


        Parameters:

          timings (dict, Optional): A dictionary containing key-value pairs
            corresponding to:

              * queuing time (in seconds)
              * sequential execution time (in seconds)
              * real speed-up obtained
              * maximum speed-up obtainable


        This method is supposed to be called only by the underlying job
        instance. It is not part of the Block's public API.

        '''

        # lock self - avoids concurrent update from scheduler/worker subsystem
        self_ = Block.objects.select_for_update().get(pk=self.pk)

        if self_.done(): return

        if self.start_date is None:
            self.start_date = self.job.start_date

        if self.job.result:

            statistics = self.job.result.stats

            info = dict(
                cpu_time = statistics.cpu['user'] + statistics.cpu['system'],
                max_memory = statistics.memory['rss'],
                data_read_size = statistics.data['volume']['read'],
                data_read_nb_blocks = statistics.data['blocks']['read'],
                data_read_time = statistics.data['time']['read'],
                data_written_size = statistics.data['volume']['write'],
                data_written_nb_blocks = statistics.data['blocks']['write'],
                data_written_time = statistics.data['time']['write'],
                stdout = self.job.result.stdout,
                stderr = self.job.result.stderr,
                error_report = self.job.result.usrerr,
                )

            if timings:
                info.update(dict(
                  queuing_time = timings['queuing'],
                  linear_execution_time = timings['linear_execution'],
                  speed_up_real = timings['speed_up_real'],
                  speed_up_maximal = timings['speed_up_maximal'],
                  ))

            self.outputs.update(**info)

        if self.job.status == Block.SKIPPED:
            self.status = Block.CACHED
        else:
            self.status = self.job.status

        if self.job.done():
            self.end_date = self.job.end_date
            r = self.job.result
            self.job.delete()
            if r: r.delete()

        # Loads Results from cache
        if self.job.result and self.analyzer and self.status == Block.CACHED:
            cache = self.first_cache()
            data_source = beat.core.data.CachedDataSource()
            data_source.setup(os.path.join(settings.CACHE_ROOT,
                    beat.core.hash.toPath(cache.hash)), settings.PREFIX)
            output_data = data_source.next()[0]
            if output_data is not None:
                algorithm = beat.core.algorithm.Algorithm(settings.PREFIX,
                        self.algorithm.fullname())
                for field, value in output_data.as_dict().items():
                    res, _ = Result.objects.get_or_create(name=field,
                        cache=cache)
                    res.primary = algorithm.results[field]['display']
                    res.type = algorithm.results[field]["type"]

                    if res.type in ['int32', 'float32', 'bool', 'string']:
                        res.data_value = str(value)
                    else:
                        res.data_value = simplejson.dumps(value, indent=4,
                            cls=NumpyJSONEncoder)

                    res.save()

            data_source.close()

        self.save()
        self._cascade_updates()
        self.experiment._update_state()


#----------------------------------------------------------


class CachedFileManager(models.Manager):

    def get_by_natural_key(self, hash):
        return self.get(hash=hash)


class CachedFile(models.Model):

    blocks = models.ManyToManyField(Block, related_name='outputs', blank=True)
    hash  = models.CharField(max_length=64, unique=True)

    # the total amount of time this block took to run considering the
    # wall-clock time.
    linear_execution_time = models.FloatField(default=0.)

    # the real speed-up obtained by running this block using X slots
    speed_up_real = models.FloatField(default=0.)

    # the maximum obtainable speed-up that could be achieved if all slots
    # were running in parallel. Essentially linear_execution_time /
    # maximum_slot_time
    speed_up_maximal = models.FloatField(default=0.)

    # the time this block waited to be executed
    queuing_time = models.FloatField(default=0.)

    stdout = models.TextField(null=True, blank=True)
    stderr = models.TextField(null=True, blank=True)
    error_report = models.TextField(null=True, blank=True)

    # other statistics of interest
    cpu_time = models.FloatField(default=0.)
    max_memory = models.BigIntegerField(default=0)
    data_read_size = models.BigIntegerField(default=0)
    data_read_nb_blocks = models.IntegerField(default=0)
    data_read_time = models.FloatField(default=0.)
    data_written_size = models.BigIntegerField(default=0)
    data_written_nb_blocks = models.IntegerField(default=0)
    data_written_time = models.FloatField(default=0.)

    objects = CachedFileManager()


    def __str__(self):
        return 'CachedFile(%s, %d blocks)' % (self.hash, self.blocks.count())


    def natural_key(self):
        return (self.hash,)


    def path(self):
        '''Returns the full path prefix to the cached file on disk'''

        return beat.core.hash.toPath(self.hash, suffix='')


    def absolute_path(self, cache=settings.CACHE_ROOT):
        '''Returns the full path prefix to the cached file on disk'''

        return os.path.join(cache, self.path())


    def files(self, cache=settings.CACHE_ROOT):
        '''Checks if any file belonging to this cache exist on disk'''

        return glob.glob(self.absolute_path(cache) + '*')


    def exists(self, cache=settings.CACHE_ROOT):
        '''Checks if any file belonging to this cache exist on disk'''

        return bool(self.files(cache))


    def index_checksums(self, cache=settings.CACHE_ROOT):
        '''Checks if this cached file indexes checksum properly'''

        abs_path = self.absolute_path(cache)
        index = sorted(glob.glob(abs_path + '*.index'))
        chksum = sorted(glob.glob(abs_path + '*.index.checksum'))

        if len(index) != len(chksum):
            logger.warn("Number of index files (%d) is different from " \
                "checksums (%d) for cache `%s'", len(index), len(chksum),
                abs_path)
            return False

        for i, c in zip(index, chksum):
            with open(c, 'rt') as f: recorded = f.read().strip()
            actual = beat.core.hash.hashFileContents(i)
            if actual != recorded:
                logger.warn("Checksum for index of cache `%s' does not " \
                    "match for file `%s' (%s != %s)", abs_path, i,
                    actual, recorded)
                return False

        return True


#----------------------------------------------------------


class BlockInputManager(models.Manager):

    def get_by_natural_key(self, name, experiment_author,
        toolchain_author, toolchain_name,
        toolchain_version, experiment_name, cache_hash, database_hash):
        block = Block.objects.get_by_natural_key(name, experiment_author,
            toolchain_author, toolchain_name, toolchain_version,
            experiment_name)
        if cache_hash:
            return self.get(cache__hash=cache_hash, block=block)
        else:
            return self.get(database__hash=database_hash, block=block)


class BlockInput(models.Model):

    block = models.ForeignKey(Block, related_name='inputs', null=True,
        on_delete=models.CASCADE)

    # if the input cames from another block, then this one is set
    cache = models.ForeignKey(CachedFile, related_name='inputs', null=True,
        on_delete=models.CASCADE)

    # if the input cames from a dataset, then this one is set
    database = models.ForeignKey(DatabaseSetOutput, related_name='blocks',
        null=True, on_delete=models.CASCADE)

    channel = models.CharField(max_length=200, default='', blank=True,
        help_text="Synchronization channel within the toolchain")

    objects = BlockInputManager()

    def natural_key(self):
        cache_hash = self.cache and self.cache.hash
        database_hash = self.database and self.database.hash
        return self.block.natural_key() + (cache_hash, database_hash)


#----------------------------------------------------------


class ResultManager(models.Manager):

    def get_by_natural_key(self, name, hash):
        return self.get(
            name=name,
            cache__hash=hash,
            )


class Result(models.Model):

    SIMPLE_TYPE_NAMES  = ('int32', 'float32', 'bool', 'string')

    cache = models.ForeignKey(CachedFile, related_name='results', null=True,
        on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    type = models.CharField(max_length=200)
    primary = models.BooleanField(default=False)
    data_value = models.TextField(null=True, blank=True)

    objects = ResultManager()


    #_____ Meta parameters __________

    class Meta:
        unique_together = ('cache', 'name')


    def __str__(self):
        return '%s - %s' % (self.cache, self.name)


    def natural_key(self):
        return (
            self.name,
            self.cache.hash,
            )


    def value(self):
        if self.data_value in ['+inf', '-inf', 'NaN']:
            return self.data_value
        elif self.type == 'int32':
            return int(self.data_value)
        elif self.type == 'float32':
            return float(self.data_value)
        elif self.type == 'bool':
            return bool(self.data_value)
        elif self.type == 'string':
            return str(self.data_value)
        elif valid_result_type(self.type):
            return simplejson.loads(self.data_value)

        return None


    def is_chart(self):

        return self.type not in Result.SIMPLE_TYPE_NAMES


def valid_result_type(value_type):
    return value_type.startswith('%s/' % settings.PLOT_ACCOUNT) or (value_type in Result.SIMPLE_TYPE_NAMES)
