#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from rest_framework import serializers

from ..common.serializers import ShareableSerializer
from ..common.fields import JSONSerializerField
from ..ui.templatetags.markup import restructuredtext

from .models import Experiment, Block

from datetime import datetime

import simplejson as json


#----------------------------------------------------------


class ExperimentSerializer(ShareableSerializer):
    name = serializers.CharField(source="fullname")
    short_name = serializers.CharField(source="name")
    toolchain = serializers.CharField(source="toolchain.fullname")
    datasets = serializers.SerializerMethodField()
    duration = serializers.SerializerMethodField()
    status = serializers.CharField(source="get_status_display")
    attestation_number = serializers.IntegerField(source="attestation.number")
    attestation_locked = serializers.BooleanField(source="attestation.locked")
    cpu_time = serializers.SerializerMethodField()
    data_read = serializers.SerializerMethodField()
    data_written = serializers.SerializerMethodField()
    description = serializers.CharField()
    declaration = JSONSerializerField()

    class Meta:
        model = Experiment
        default_fields = ShareableSerializer.Meta.default_fields + ['name', 'toolchain',
                  'datasets', 'short_description',
                  'creation_date', 'start_date', 'end_date',
                  'duration', 'status',
                  'accessibility',
                  'attestation_number', 'attestation_locked',
                  'cpu_time', 'data_read', 'data_written',
                  'is_owner']


    def __init__(self, *args, **kwargs):
        super(ExperimentSerializer, self).__init__(*args, **kwargs)
        self.statistics = {}

    def __get_statistics(self, obj, key):
        if not obj in self.statistics:
            self.statistics[obj] = {
                'cpu_time': 0,
                'data_read': 0,
                'data_written': 0
            }
            for block in obj.blocks.all():
                self.statistics[obj]['cpu_time'] += block.cpu_time()
                self.statistics[obj]['data_read'] += block.data_read_size()
                self.statistics[obj]['data_written'] += \
                    block.data_written_size()
        return self.statistics[obj][key]

    def get_datasets(self, obj):
        return map(lambda d: d.fullname(), obj.referenced_datasets.all())

    def get_start_date(self, obj):
        if obj.is_busy():
            return None
        elif obj.is_done() or obj.is_running():
            return obj.start_date
        else:
            return obj.creation_date

    def get_end_date(self, obj):
        if obj.is_busy():
            return None
        elif obj.is_done():
            return obj.end_date
        else:
            return datetime.now()

    def get_duration(self, obj):
        start_date = self.get_start_date(obj)
        end_date = self.get_end_date(obj)

        if (start_date is None) or (end_date is None):
            return '-'

        duration = end_date - start_date

        duration_seconds = max(duration.seconds, 1) # At least one second

        seconds = duration_seconds % 60
        minutes = ((duration_seconds - seconds) % 3600) / 60
        hours = duration.days * 24 + (duration_seconds - minutes * 60 - seconds) / 3600

        if seconds > 0:
            minutes += 1

            if minutes == 60:
                minutes = 0
                hours += 1

        return '%dh%02d' % (hours, minutes)

    def get_cpu_time(self, obj):
        return self.__get_statistics(obj, 'cpu_time')

    def get_data_read(self, obj):
        return self.__get_statistics(obj, 'data_read')

    def get_data_written(self, obj):
        return self.__get_statistics(obj, 'data_written')


#----------------------------------------------------------


class AnalyzerSerializer(serializers.ModelSerializer):
    results = serializers.SerializerMethodField()

    class Meta:
        model = Block

    def get_results(self, obj):
        results = {}

        db_results = obj.results.all()
        if db_results.count() > 0:
            for result_entry in db_results:
                results[result_entry.name] = {
                    'type':    result_entry.type,
                    'primary': result_entry.primary,
                    'value':   result_entry.value(),
                }
        else:
            dataformat_declaration = json.loads(obj.algorithm.result_dataformat)

            for field, type in dataformat_declaration.items():
                primary = (field[0] == '+')

                results[field[1:] if primary else field] = {
                    'type':    type,
                    'primary': primary,
                    'value':   None,
                }

        return results


#----------------------------------------------------------


class BlockErrorSerializer(serializers.ModelSerializer):
    block = serializers.CharField(source='name')
    algorithm = serializers.CharField(source='algorithm.fullname')
    stdout = serializers.CharField()
    stderr = serializers.CharField()
    details = serializers.CharField(source='error_report')

    class Meta:
        model = Block


#----------------------------------------------------------


class ExperimentResultsSerializer(ShareableSerializer):
    started = serializers.SerializerMethodField()
    done = serializers.SerializerMethodField()
    failed = serializers.SerializerMethodField()
    attestation = serializers.IntegerField(source='attestation.number')
    blocks_status = serializers.SerializerMethodField()
    execution_info = serializers.SerializerMethodField()
    results = serializers.SerializerMethodField()
    errors = serializers.SerializerMethodField()
    html_description = serializers.SerializerMethodField()
    description = serializers.SerializerMethodField()
    declaration = JSONSerializerField()

    class Meta(ShareableSerializer.Meta):
        model = Experiment
        default_fields = ['started', 'done', 'failed', 'blocks_status',
                          'results', 'attestation', 'declaration',
                          'errors', 'sharing', 'accessibility',
                          'execution_info']

    def get_started(self, obj):
        return obj.status not in [Experiment.PENDING, Experiment.SCHEDULED, Experiment.CANCELING]

    def get_done(self, obj):
        return obj.status in [Experiment.DONE, Experiment.FAILED]

    def get_failed(self, obj):
        return obj.status == Experiment.FAILED

    def get_blocks_status(self, obj):
        results = {}
        for block in obj.blocks.iterator():
            if block.status == Block.CACHED:
                results[block.name] = 'generated'
            elif block.status == Block.NOT_CACHED:
                results[block.name] = 'pending'
            elif block.status == Block.FAILED:
                results[block.name] = 'failed'
            else:
                results[block.name] = 'processing'
        return results

    def get_execution_info(self, obj):
        results = {}
        for block in obj.blocks.iterator():
            results[block.name] = {
                'linear_execution_time':   block.linear_execution_time(),
                'speed_up_real':           block.speed_up_real(),
                'speed_up_maximal':        block.speed_up_maximal(),
                'queuing_time':            block.queuing_time(),
            }
        return results

    def get_results(self, obj):
        results = {}
        for k in obj.blocks.filter(analyzer=True).all():
            serializer = AnalyzerSerializer(k)
            results[k.name] = serializer.data['results']
        return results

    def get_errors(self, obj):
        serializer = BlockErrorSerializer(obj.blocks.filter(outputs__error_report__isnull=False), many=True)
        return serializer.data

    def get_html_description(self, obj):
        d = obj.description.decode('utf-8')
        if len(d) > 0:
            return restructuredtext(d)
        return ''

    def get_description(self, obj):
        return obj.description.decode('utf-8')
