#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.http import Http404, HttpResponseForbidden
from django.shortcuts import get_object_or_404
from django.shortcuts import render_to_response
from django.template import RequestContext
from django.views.decorators.csrf import csrf_protect
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.conf import settings

from .models import Library
from ..common.texts import Messages
from ..team.models import Team
from ..ui.templatetags.markup import restructuredtext

from beat.core import prototypes

import simplejson as json


@csrf_protect
@login_required
def create(request):
    """Creates a new library.

    The user must be authenticated before it can add a new library
    """

    visible_libraries = Library.objects.from_author_and_public(request.user,
            request.user.username).order_by('-creation_date')

    return render_to_response('libraries/edition.html',
            dict(
                op="new",
                texts=Messages,
                visible_libraries=visible_libraries,
                ),
            context_instance=RequestContext(request)
            )


@csrf_protect
@login_required
def new_version(request, name):
    """Creates a new version of a library.

    The user must be authenticated before it can add a new library
    """

    previous_version = Library.objects.filter(author=request.user,
            name__iexact=name).order_by('-version')

    if not previous_version: raise Http404()

    visible_libraries = Library.objects.from_author_and_public(request.user,
            request.user.username).order_by('-creation_date')

    return render_to_response('libraries/edition.html',
            dict(
                op="new-version",
                texts=Messages,
                library=previous_version[0],
                visible_libraries=visible_libraries,
                ),
            context_instance=RequestContext(request)
            )


@csrf_protect
@login_required
def fork(request, author_name, name, version):
    """Creates a new library by forking an existing library

    Answers to the following URLs:

    <libraries>/fork/<author>/<name>/<version>/

    The user must be authenticated before it can fork a library
    """

    # Retrieves the forked library based on user accessible libraries
    fork_of = get_object_or_404(Library.objects.for_user(request.user, True),
                                author__username__iexact=author_name,
                                name__iexact=name,
                                version=int(version)
                               )

    # The only case a user can't fork an library is if it's usable rather
    # than shared
    if fork_of.sharing == Library.USABLE and fork_of.author != request.user:
      raise Http404()

    visible_libraries = Library.objects.from_author_and_public(request.user,
            request.user.username).order_by('-creation_date')

    return render_to_response('libraries/edition.html',
            dict(
                op="fork",
                texts=Messages,
                library=fork_of,
                visible_libraries=visible_libraries,
                ),
            context_instance=RequestContext(request))


#----------------------------------------------------------


@login_required
def edit(request, author_name, name, version):
    """Allows the modification of the library

    The user must be authenticated before it can edit a library
    """

    # Retrieves the library
    library = get_object_or_404(Library, author__username=author_name,
            name__iexact=name, version=int(version))

    # only author can edit it
    if library.author != request.user: raise Http404()

    if not library.modifiable():
        return HttpResponseForbidden('Library %s is not modifiable' % \
                library.fullname())

    visible_libraries = Library.objects.from_author_and_public(request.user,
            request.user.username).exclude(id=library.id).order_by('-creation_date')

    return render_to_response('libraries/edition.html',
            dict(
                op="edit",
                texts=Messages,
                library=library,
                visible_libraries=visible_libraries,
                ),
            context_instance=RequestContext(request))


#----------------------------------------------------------


def view(request, author_name, name, version=None):
    """Shows the library"""

    # Retrieves the library
    if version:
        library = get_object_or_404(
                Library,
                author__username__iexact=author_name,
                name__iexact=name,
                version=int(version),
                )
    else:
        library = Library.objects.filter(author__username__iexact=author_name,
                name__iexact=name).order_by('-version')
        if not library:
            raise Http404()
        else:
            library = library[0]

    (has_access, _, __) = library.accessibility_for(request.user)

    if not has_access: raise Http404()

    owner = (request.user == library.author)

    # Users the object can be shared with
    users = User.objects.exclude(username__in=settings.ACCOUNTS_TO_EXCLUDE_FROM_TEAMS).order_by('username')

    # Render the page
    return render_to_response('libraries/view.html',
                              {
                                  'library': library,
                                  'owner': owner,
                                  'users': users,
                                  'teams': Team.objects.all(),
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def diff(request, author1, name1, version1, author2, name2, version2):
    """Shows the difference between two libraries. The Web API is used to retrieve the
    details about the libraries and check the accessibility.

    Answers to the following URLs

    <libraries>/diff/<author1>/<name1>/<version1>/<author2>/<name2>/<version2>/

    """

    return render_to_response('libraries/diff.html',
                              {'library1_author':  author1,
                               'library1_name':    name1,
                               'library1_version': version1,
                               'library2_author':  author2,
                               'library2_name':    name2,
                               'library2_version': version2,
                              },
                              context_instance=RequestContext(request))


#----------------------------------------------------------


def ls(request, author_name):
    '''List all accessible libraries to the request user'''

    if not author_name: return public_ls(request)

    # check that the user exists on the system
    author = get_object_or_404(User, username=author_name)

    # orders libraries so that the latest information is displayed first
    objects = Library.objects.from_author_and_public(request.user,
            author_name).order_by('-creation_date')
    objects = Library.filter_latest_versions(objects)

    return render_to_response('libraries/list.html',
            dict(
                objects=objects,
                author=author,
                owner=(request.user==author),
                ),
            context_instance=RequestContext(request),
            )


#----------------------------------------------------------


def public_ls(request):
    '''List all publicly accessible objects'''

    # orders so that latest is displayed first
    objects = Library.objects.public().order_by('-creation_date')
    objects = Library.filter_latest_versions(objects)

    return render_to_response('libraries/list.html',
            dict(
                objects=objects,
                author=request.user, #anonymous
                owner=False,
                ),
            context_instance=RequestContext(request),
            )
