/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 *
 * This file is part of the beat.web module of the BEAT platform.
 *
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 *
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
 */

/*
 * ReportService
 * Desc:
 * 	The datastore for the reports app, including data fetched from the
 * 	server and cross-component app state
 */
angular.module('reportApp').factory('ReportService', ['$http', function($http){
	let reportServiceInstance = {};
	// experiments of reports are in arbitrary groups,
	// in a many-to-many relationship
	let groupData = [];

	// represents a Group in the report
	// has a name and a list of experiments that belong to it
	class Group {
		constructor (name) {
			this._name = name;
			this._analyzer = '';
			this._experimentNames = new Set();
			this._reportItemIds = new Set();
		}

		// get the experiment names in this group
		get experiments () {
			return Array.from(this._experimentNames);
		}

		// get the ids of report items that use this group
		get reportItemIds () {
			return Array.from(this._reportItemIds);
		}

		// get the group name
		get name () {
			return this._name;
		}

		// get the analyzer of the experiments in this group
		get analyzer () {
			return this._analyzer;
		}

		set analyzer (analyzer) {
			console.log(`setting anayzler = ${analyzer}`);
			this._analyzer = analyzer;
		}

		// add an exp to this group
		addExperiment (expName) {
			return this._experimentNames.add(expName);
		}

		// rm an exp from this group
		removeExperiment (expName) {
			return this._experimentNames.delete(expName);
		}

		// add an exp to this group
		addReportItem (id) {
			return this._reportItemIds.add(id);
		}

		// rm an exp from this group
		removeReportItem (id) {
			return this._reportItemIds.delete(id);
		}
	};

	// gets groups
	reportServiceInstance.groups = groupData;
	// TODO: move this hacky state to the addReportItem miniwindows
	reportServiceInstance.activeGroup = '';

	// serializes groups as an object with form:
	// {
	// 	<group name 1>: { experiments: [], reportItemIds: [], analyzer: '' },
	// 	...
	// }
	reportServiceInstance.serializeGroups = () => {
		return groupData
		.map(g => { return {
			[g.name]: {
				experiments: g.experiments,
				reportItemIds: g.reportItemIds,
				analyzer: g.analyzer
			}
		};
		})
		.reduce((o, g) => Object.assign(o, g), {});
	};

	// create a new group for the report
	// returns false if it already exists
	// returns the newly added group if successful
	reportServiceInstance.createGroup = (name) => {
		if(typeof name !== 'string'){
			throw new Error(`new group name is not a string: ${JSON.stringify(name)}`);
		}

		if(groupData.find(g => g.name === name)){
			return false;
		}

		let g = new Group(name);

		groupData.push(g);
		// change active group to newest group by default
		reportServiceInstance.activeGroup = g.name;
		return g;
	};

	// delete a group
	reportServiceInstance.deleteGroup = (name) => {
		let idx = groupData.indexOf(groupData.find(g => g.name === name));
		if (idx > -1) {
			groupData.splice(idx, 1);
		}
	};

	// load group info from the serialized format:
	// {
	// 	<group name 1>: { experiments: [], reportItemIds: [], analyzer: '' },
	// 	...
	// }
	reportServiceInstance.loadGroups = (data) => {
		// wipe data and load groups
		groupData.splice(0, groupData.length);

		// if there's no group data, create a default one
		if(!data || Object.keys(data).length === 0){
			reportServiceInstance.createGroup('default');
			return;
		}

		Object.entries(data)
		// sometimes we get an empty string for name for some reason
		.filter(([groupName, gData]) => groupName.length > 0)
		.forEach(([groupName, gData]) => {
			let g = reportServiceInstance.createGroup(groupName);
			g.analyzer = gData.analyzer;
			gData.experiments.forEach(n => g.addExperiment(n));
			gData.reportItemIds.forEach(id => g.addReportItem(id));
		});
	};

	// save group info via sending to server
	reportServiceInstance.saveGroups = () => {
		console.warn(`not implemented: reportServiceInstance.saveGroups`);
	};

	// add experiment to group
	reportServiceInstance.addExperimentToGroup = (expName, groupName, analyzer) => {
		checkForGroup(groupName);

		let group = groupData.find(g => g.name === groupName);
		return group.addExperiment(expName, analyzer);
	};

	// rm experiment from group
	reportServiceInstance.removeExperimentFromGroup = (expName, groupName) => {
		checkForGroup(groupName);
		let group = groupData.find(g => g.name === groupName);
		return group.removeExperiment(expName);
	};

	// gets experiments from a group
	reportServiceInstance.getGroupExperiments = (groupName) => {
		checkForGroup(groupName);
		let group = groupData.find(g => g.name === groupName);
		return group.experiments;
	};

	// gets groups for an experiment
	reportServiceInstance.getExperimentGroups = (expName) => {
		return groupData.filter(g => g.experiments.includes(expName)).map(g => g.name);
	};

	// add report item to group
	reportServiceInstance.addReportItemToGroup = (id, groupName) => {
		checkForGroup(groupName);

		let group = groupData.find(g => g.name === groupName);
		group.addReportItem(id);
	};

	// rm report item id from group
	reportServiceInstance.removeReportItemFromGroup = (id, groupName) => {
		checkForGroup(groupName);
		let group = groupData.find(g => g.name === groupName);
		return group.removeReportItem(id);
	};

	// gets group for a report item's id
	reportServiceInstance.getReportItemGroup = (id) => {
		return groupData.find(g => g.reportItemIds.includes(id));
	};

	// helper to assert that a group exists
	function checkForGroup (groupName) {
		if(!groupData.find(g => g.name === groupName)){
			throw new Error(`Could not find group "${JSON.stringify(groupName)}"`);
		}
	}

	return reportServiceInstance;
}]);
