#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.contrib import admin
from django import forms

from ..common.texts import Messages

from .models import Search, Leaderboard, Rank
from ..ui.forms import CodeMirrorRSTCharField
from ..ui.forms import CodeMirrorJSONCharField
from ..ui.forms import NameField


#----------------------------------------------------------


def rehash_search(modeladmin, request, queryset):
    """Recalculates the hash of a search"""
    for q in queryset:
        q.save()

rehash_search.short_description = 'Rehash selected search'


#----------------------------------------------------------


class SearchModelForm(forms.ModelForm):

    name = NameField(
            widget=forms.TextInput(attrs=dict(size=80)),
            help_text=Messages['algo_name'],
            )

    description = CodeMirrorRSTCharField(
            required=False,
            help_text=Messages['description'],
            )

    filters = CodeMirrorJSONCharField(
            readonly=False,
            required=False,
            help_text=Messages['json'],
            )

    settings = CodeMirrorJSONCharField(
            readonly=False,
            required=False,
            help_text=Messages['json'],
            )

    class Meta:
        model = Search
        exclude = []
        widgets = {
                'short_description': forms.TextInput(
                    attrs=dict(size=100),
                    ),
                }


#----------------------------------------------------------


class SearchAdmin(admin.ModelAdmin):

    list_display        = ('id', 'author', 'name')
    search_fields       = [
                            'author__username',
                            'name',
                            'short_description',
                            'description',
                            'filters',
                            'settings',
                          ]
    list_display_links  = ('id', 'name')

    list_filter         = ('author', 'name', 'version')
    readonly_fields    = ('hash',)

    actions = [
        rehash_search,
    ]

    form = SearchModelForm

    filter_horizontal = [
        'shared_with',
        'shared_with_team'
    ]

    fieldsets = (
        (None,
          dict(
            fields=('name', 'author',),
            ),
          ),
        ('Documentation',
          dict(
            classes=('collapse',),
            fields=('short_description', 'description',),
            ),
          ),
        ('Versioning',
          dict(
            classes=('collapse',),
            fields=('version', 'previous_version', 'fork_of'),
            ),
          ),
        ('Sharing',
          dict(
            classes=('collapse',),
            fields=('sharing', 'shared_with', 'shared_with_team'),
            ),
          ),
        ('Definition',
          dict(
            fields=('hash', 'filters', 'settings'),
            ),
          ),
        )

admin.site.register(Search, SearchAdmin)


class RankInline(admin.TabularInline):
    model           = Rank
    can_delete      = False
    extra           = 0
    max_num         = 0
    readonly_fields = ('id', 'algorithm', 'result', 'order', 'experiment')
    ordering        = ('algorithm', 'order',)


def number_of_experiments(obj):
    return obj.experiments.count()
number_of_experiments.short_description = 'Experiments'


def users_to_notify(obj):
    return obj.notify.count()
users_to_notify.short_description = 'Subscribed'


def search_sharing(obj):
    return obj.search.get_sharing_display()
search_sharing.short_description = 'Sharing'


class LeaderboardAdmin(admin.ModelAdmin):

    list_display = ('id', 'search', 'created', 'updated', users_to_notify, number_of_experiments, search_sharing)
    search_fiels = [
                      'search__author__username',
                      'search__name',
                      'search_short_description',
                      'search_description',
                      'filters',
                      'settings',
                      ]
    list_display_links = ('id', 'search',)
    inlines = [
        RankInline,
        ]

    filter_horizontal = [
        'notify',
        ]

admin.site.register(Leaderboard, LeaderboardAdmin)
