#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################


import datetime
import itertools

import simplejson

from django.shortcuts import render_to_response
from django.template import RequestContext
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseForbidden
from django.db.models import Sum

from .models import HourlyStatistics

def calculate_totals():
    """Caculates all totals required by the statistics display"""

    from django.contrib.auth.models import User
    from ..databases.models import Database
    from ..backend.models import Environment, Queue, Worker
    from ..experiments.models import Experiment
    from ..toolchains.models import Toolchain
    from ..algorithms.models import Algorithm
    from ..libraries.models import Library
    from ..dataformats.models import DataFormat
    from ..team.models import Team
    from ..attestations.models import Attestation
    from ..reports.models import Report
    from ..plotters.models import Plotter
    from ..search.models import Search

    # for calculating the total cpu time, we use the HourlyStatistics and
    # accumulate over the whole history
    objects = HourlyStatistics.objects.all()
    counter = objects.count()

    details = objects.aggregate(
            cpu_time=Sum('cpu_time'),
            max_memory=Sum('max_memory'),
            data_read_size=Sum('data_read_size'),
            data_read_time=Sum('data_read_time'),
            data_written_size=Sum('data_written_size'),
            data_written_time=Sum('data_written_time'),
            )

    cpu_time = details['cpu_time']
    memory = details['max_memory']
    input_size = details['data_read_size']
    input_time = details['data_read_time']
    output_size = details['data_written_size']
    output_time = details['data_written_time']

    return dict(

            counter=counter,
            cpu_time_hours=int(cpu_time/float(60**2)) if cpu_time else 0,
            memory_gb=int(memory/float(2**30)) if memory else 0,
            input_gb = int(input_size/float(2**30)) if input_size else 0,
            input_bw_mb_s = int((input_size/float(2**20))/input_time) if input_size else 0,
            output_gb = int(output_size/float(2**30)) if output_size else 0,
            output_bw_mb_s = int((output_size/float(2**20))/output_time) if output_size else 0,

            users=User.objects.count(),
            databases=Database.objects.count(),
            environments=Environment.objects.count(),
            queues=Queue.objects.count(),
            workers=Worker.objects.count(),

            experiments=Experiment.objects.count(),
            toolchains=Toolchain.objects.count(),
            algorithms=Algorithm.objects.count(),
            libraries=Library.objects.count(),
            dataformats=DataFormat.objects.count(),
            teams=Team.objects.count(),

            attestations=Attestation.objects.count(),
            searches=Search.objects.count(),
            reports=Report.objects.count(),
            plotters=Plotter.objects.count(),

            )


def convert_values(stats):
    """Converts values so they are suitable for plotting"""

    # transform into plottable data
    def _seconds_to_hours(s, field):
        if s is not None: return s[field]/float(60**2) #hours
        return 0.

    def _bytes_to_gb(s, field):
        if s is not None: return s[field]/float(2**30) #gigabytes
        return 0.

    def _bw_in_mbs(s, num, den):
        if s is None: return 0.
        if not s[den]: return 0. #avoid division by zero
        return (s[num]/float(2**20))/s[den] #mb per sec

    return dict(
            cpu_time=[_seconds_to_hours(k, 'cpu_time') for k in stats],
            memory=[_bytes_to_gb(k, 'max_memory') for k in stats],
            input_bw=[_bw_in_mbs(k, 'data_read_size', 'data_read_time') for k in stats],
            output_bw=[_bw_in_mbs(k, 'data_written_size', 'data_written_time') for k in stats],
            )


def grouper(n, iterable, fillvalue=None):
    "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.izip_longest(fillvalue=fillvalue, *args)


def get_statistics(hours_to_go_back, cluster_by):
    """Returns the data for hourly charts on various axes"""

    # entries for X days, including today
    now = datetime.datetime.now()
    back = now - datetime.timedelta(hours=hours_to_go_back)
    available = HourlyStatistics.objects.filter(date__gt=back).order_by('date',
            'hour')

    # this must be true!
    assert available.count() <= hours_to_go_back

    # annotate real entries with the actual timestamp
    onehour = datetime.timedelta(hours=1)
    start = back + onehour

    def _hours_since_back(o):
        timestamp = datetime.datetime.combine(o.date, datetime.time(o.hour))
        return int((timestamp - back).total_seconds()/(60**2))

    annotated = dict([(_hours_since_back(k), k) for k in available])

    # complete the holes as there must be X entries available
    stats = hours_to_go_back * [None]
    for k in annotated: stats[k] = annotated[k]

    # cluster using the amount of hours provided by ``cluster_by``
    groups = grouper(cluster_by, stats)

    # calculate totals for all entries defined
    def _totals(group, fields):
        retval = {}
        for f in fields:
            retval[f] = sum([getattr(k, f) for k in group if k is not None])
        return retval

    fields = ('cpu_time', 'max_memory', 'data_read_size', 'data_read_time',
            'data_written_size', 'data_written_time')
    summaries = [_totals(g, fields) for g in groups]

    # setup data so information is shown for every field accros groups
    return convert_values(summaries)


def hourly_charts():
    """Returns the data for hourly charts on various axes"""

    # a set of intuitive labels for all data points
    labels = 24*['']
    labels[-1] = 'now'
    labels[17] = '-6h'
    labels[11] = '-12h'
    labels[5] = '-18h'
    labels[0] = '-23h'

    retval = dict(labels=labels)
    retval.update(get_statistics(hours_to_go_back=24, cluster_by=1))
    return simplejson.dumps(retval)


def daily_charts():
    """Returns the data for daily charts on various axes"""

    # a set of intuitive labels for all data points
    labels = 30*['']
    labels[-1] = 'today'
    labels[22] = '-7d'
    labels[14] = '-15d'
    labels[7] = '-22d'
    labels[0] = '-29d'

    retval = dict(labels=labels)
    retval.update(get_statistics(hours_to_go_back=24*30, cluster_by=24))
    return simplejson.dumps(retval)


def weekly_charts():
    """Returns the data for weekly charts on various axes"""

    # a set of intuitive labels for all data points
    labels = 24*['']
    labels[-1] = 'this week'
    labels[17] = '-6w'
    labels[11] = '-12w'
    labels[5] = '-18w'
    labels[0] = '-23w'

    retval = dict(labels=labels)
    retval.update(get_statistics(hours_to_go_back=24*7*24,
        cluster_by=24*7))
    return simplejson.dumps(retval)


@login_required
def statistics(request):
    """Displays a summary of statistics as charts

    A couple of information blocks will be displayed at the top of the page,
    summarizing processing information and the total number of available
    objects, for informational purposes.

    Furthermore, we will have a tabbed view following the summary block,
    containing the following charts (one for cpu-time, one for real/wall-clock
    processing time, maximum virtual memory, data in/output with number of
    bytes read/written and bandwidth):

    * The hourly display will show the last 24 hours of activity
    * The daily display will show the last 7 days of activity
    * The weekly display will show the last 6 months (24 weeks) of activity

    """

    if not(request.user.is_superuser): return HttpResponseForbidden()

    return render_to_response('statistics/statistics.html',
            dict(
                totals=calculate_totals(),
                hourly_chart_data=hourly_charts(),
                daily_chart_data=daily_charts(),
                weekly_chart_data=weekly_charts(),
                ),
            context_instance=RequestContext(request))
