#!/usr/bin/env python
# vim: set fileencoding=utf-8 :

###############################################################################
#                                                                             #
# Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/           #
# Contact: beat.support@idiap.ch                                              #
#                                                                             #
# This file is part of the beat.web module of the BEAT platform.              #
#                                                                             #
# Commercial License Usage                                                    #
# Licensees holding valid commercial BEAT licenses may use this file in       #
# accordance with the terms contained in a written agreement between you      #
# and Idiap. For further information contact tto@idiap.ch                     #
#                                                                             #
# Alternatively, this file may be used under the terms of the GNU Affero      #
# Public License version 3 as published by the Free Software and appearing    #
# in the file LICENSE.AGPL included in the packaging of this file.            #
# The BEAT platform is distributed in the hope that it will be useful, but    #
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY  #
# or FITNESS FOR A PARTICULAR PURPOSE.                                        #
#                                                                             #
# You should have received a copy of the GNU Affero Public License along      #
# with the BEAT platform. If not, see http://www.gnu.org/licenses/.           #
#                                                                             #
###############################################################################

from django.db import models
from django.db.models import Q
from django.core.urlresolvers import reverse
from django.contrib.auth.models import User
from django.utils.translation import ugettext_lazy as _


class TeamManager(models.Manager):

    def get_by_natural_key(self, owner_username, name):
        return self.get(owner__username=owner_username, name=name)

    def for_user(self, user, add_public=False):

        if user.is_anonymous():
            return self.public()

        query = Q(owner=user) |\
                (Q(privacy_level__in=[Team.MEMBERS, Team.PRIVATE]) & Q(members=user))

        if add_public:
            query |= Q(privacy_level=Team.PUBLIC)

        return self.filter(query).distinct().order_by('privacy_level')

    def public(self):
        return self.filter(privacy_level=Team.PUBLIC)


class Team(models.Model):

    #_____ Constants __________

    PRIVATE = 0
    MEMBERS = 1
    PUBLIC  = 2


    #_____ Fields __________

    PRIVACY_LEVELS = [
        [PRIVATE, _('Private')],
        [MEMBERS, _('Members only')],
        [PUBLIC,  _('Public')],
    ]

    owner             = models.ForeignKey(User)
    name              = models.CharField(max_length=32)
    short_description = models.TextField(max_length=100, default='', blank=True)
    members           = models.ManyToManyField(User, related_name='teams', blank=True)
    privacy_level     = models.PositiveIntegerField(default=PRIVATE, choices=PRIVACY_LEVELS)

    objects = TeamManager()


    #_____ Meta parameters __________

    class Meta:
        unique_together = ['owner', 'name']


    #_____ Utilities __________

    def natural_key(self):

        return (self.owner.username, self.name)


    def total_shares(self, user=None):
        '''Count number of objects shared with this team

        If ``user`` is passed, counts objects shared with this team that the
        given user has access to.
        '''

        used_at = 0

        if user is None:

            used_at += self.shared_algorithms.count()
            used_at += self.shared_databases.count()
            used_at += self.shared_dataformats.count()
            used_at += self.shared_environments.count()
            used_at += self.shared_experiments.count()
            used_at += self.shared_librarys.count()
            used_at += self.shared_plotterparameters.count()
            used_at += self.shared_plotters.count()
            used_at += self.shared_searchs.count()
            used_at += self.shared_toolchains.count()
            used_at += self.usable_algorithms.count()
            used_at += self.usable_librarys.count()
            used_at += self.usable_plotterparameters.count()
            used_at += self.usable_plotters.count()

        else:

            used_at += self.shared_algorithms.for_user(user, add_public=True).count()
            used_at += self.shared_databases.for_user(user, add_public=True).count()
            used_at += self.shared_dataformats.for_user(user, add_public=True).count()
            used_at += self.shared_environments.for_user(user, add_public=True).count()
            used_at += self.shared_experiments.for_user(user, add_public=True).count()
            used_at += self.shared_librarys.for_user(user, add_public=True).count()
            used_at += self.shared_plotterparameters.for_user(user, add_public=True).count()
            used_at += self.shared_plotters.for_user(user, add_public=True).count()
            used_at += self.shared_searchs.for_user(user, add_public=True).count()
            used_at += self.shared_toolchains.for_user(user, add_public=True).count()
            used_at += self.usable_algorithms.for_user(user, add_public=True).count()
            used_at += self.usable_librarys.for_user(user, add_public=True).count()
            used_at += self.usable_plotterparameters.for_user(user, add_public=True).count()
            used_at += self.usable_plotters.for_user(user, add_public=True).count()

        return used_at


    def deletable(self):

        return not self.total_shares()


    #_____ Overrides __________

    def __unicode__(self):
        return self.fullname()


    #_____ Methods __________

    def fullname(self):
        return '%s/%s' % (self.owner.username, self.name)


    def get_absolute_url(self):
        return reverse('teams:view', args=(self.owner.username, self.name))
