/*
 * Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
 * Contact: beat.support@idiap.ch
 * 
 * This file is part of the beat.web module of the BEAT platform.
 * 
 * Commercial License Usage
 * Licensees holding valid commercial BEAT licenses may use this file in
 * accordance with the terms contained in a written agreement between you
 * and Idiap. For further information contact tto@idiap.ch
 * 
 * Alternatively, this file may be used under the terms of the GNU Affero
 * Public License version 3 as published by the Free Software and appearing
 * in the file LICENSE.AGPL included in the packaging of this file.
 * The BEAT platform is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * You should have received a copy of the GNU Affero Public License along
 * with the BEAT platform. If not, see http://www.gnu.org/licenses/.
*/
/* beat.toolchains.common.js

   Declaration of the classes and constants common to the Toolchain Editor and the
   Toolchain Viewer panels
*/


// Declaration of our namespaces
if (beat === undefined)
    var beat = {}

if (beat.toolchains === undefined)
    beat.toolchains = {}

if (beat.toolchains.common === undefined)
    beat.toolchains.common = {}


/*************************************** CONSTANTS **************************************/

beat.toolchains.common.CELL_SPACING = 30;

beat.toolchains.common.BLOCK_DEFAULT_WIDTH   = 6;
beat.toolchains.common.DATASET_DEFAULT_WIDTH = 5;
beat.toolchains.common.BLOCK_DEFAULT_HEIGHT  = 3;
beat.toolchains.common.BLOCK_CHANNEL_RADIUS  = 8;

beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT       = 17;
beat.toolchains.common.BLOCK_CONNECTOR_CENTER       = (beat.toolchains.common.BLOCK_CONNECTOR_HEIGHT + 3) / 2;
beat.toolchains.common.BLOCK_CONNECTOR_HANDLE_WIDTH = 7;

// Connection selection state
beat.toolchains.common.SELECTION_NONE   = 0;
beat.toolchains.common.SELECTION_INPUT  = 1;
beat.toolchains.common.SELECTION_OUTPUT = 2;
beat.toolchains.common.SELECTION_FULL   = 3;



/*********************************** UTILITY FUNCTIONS **********************************/

//----------------------------------------------------------------------------------------
// Convert a value from [cells] to [pixels]
//----------------------------------------------------------------------------------------
beat.toolchains.common.cellsToPixels = function(value_in_cells, zoom)
{
    return Math.floor(value_in_cells * beat.toolchains.common.CELL_SPACING * zoom);
}


//----------------------------------------------------------------------------------------
// Convert a value from [pixels] to [cells]
//----------------------------------------------------------------------------------------
beat.toolchains.common.pixelsToCells = function(value_in_pixels, zoom)
{
    return Math.floor(value_in_pixels / (beat.toolchains.common.CELL_SPACING * zoom));
}


//----------------------------------------------------------------------------------------
// Convert a value from [pixels] to the nearest one in [cells]
//----------------------------------------------------------------------------------------
beat.toolchains.common.nearestCell = function(value_in_pixels, zoom)
{
    var value_in_cells = beat.toolchains.common.pixelsToCells(value_in_pixels, zoom);

    if (value_in_pixels - beat.toolchains.common.cellsToPixels(value_in_cells, zoom) >
            0.5 * beat.toolchains.common.CELL_SPACING * zoom)
    {
        ++value_in_cells;
    }

    return value_in_cells;
}


/***************************** CLASS: BlockRepresentation *******************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.common.BlockRepresentation = function(block)
{
    this.block                    = block;
    this.left                     = 0;   // In cells or in pixels (see 'coords_in_pixels')
    this.top                      = 0;
    this.width                    = (block.isDataset() ? beat.toolchains.common.DATASET_DEFAULT_WIDTH
                                                       : beat.toolchains.common.BLOCK_DEFAULT_WIDTH);
    this.height                   = Math.max(block.nbInputs(), block.nbOutputs()) + 2;
    this.selected                 = false;
    this.coords_in_pixels         = false;
    this.input_coords             = [];
    this.output_coords            = [];
    this.add_input_button_coords  = null;
    this.add_output_button_coords = null;
    this.highlighted_button       = null; // -1: add_input, -2: add_output,
                                          // 0 -> N: inputs, N+1 -> M: outputs

    for (var i = 0; i < block.nbInputs(); ++i)
        this.input_coords.push(null);

    for (var i = 0; i < block.nbOutputs(); ++i)
        this.output_coords.push(null);
}


//----------------------------------------------------------------------------------------
// Returns the type of the component
//----------------------------------------------------------------------------------------
beat.toolchains.common.BlockRepresentation.prototype.type = function()
{
    return (this.block.isAnalyzer() ? beat.toolchains.editor.TYPE_ANALYZER
                                    : (this.block.isDataset() ? beat.toolchains.editor.TYPE_DATASET
                                                              : beat.toolchains.editor.TYPE_BLOCK));
}


//----------------------------------------------------------------------------------------
// Returns the type of the component
//----------------------------------------------------------------------------------------
beat.toolchains.common.BlockRepresentation.prototype.getBoundingBoxInPixels = function(zoom)
{
    if (this.coords_in_pixels)
    {
        return {
            left:   this.left,
            top:    this.top,
            right:  this.left + beat.toolchains.common.cellsToPixels(this.width, zoom),
            bottom: this.top + beat.toolchains.common.cellsToPixels(this.height, zoom),
        };
    }
    else
    {
        return {
            left:   beat.toolchains.common.cellsToPixels(this.left, zoom),
            top:    beat.toolchains.common.cellsToPixels(this.top, zoom),
            right:  beat.toolchains.common.cellsToPixels(this.left + this.width, zoom),
            bottom: beat.toolchains.common.cellsToPixels(this.top + this.height, zoom),
        };
    }
}


/************************** CLASS: ConnectionRepresentation *****************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation = function(connection)
{
    this.connection           = connection;
    this.points               = [];
    this.selected             = beat.toolchains.common.SELECTION_NONE;
    this.active_segment       = null;
    this.active_segment_index = null;
    this.translation_data     = null;

    this.bounding_box = {
        left:   0,
        top:    0,
        right:  0,
        bottom: 0,
    };
}


//----------------------------------------------------------------------------------------
// Returns the type of the component
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.type = function()
{
    return beat.toolchains.editor.TYPE_CONNECTION;
}


//----------------------------------------------------------------------------------------
// Returns the number of segments
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.nbSegments = function()
{
    if (this.points.length < 2)
        return 0;

    return this.points.length - 1;
}


//----------------------------------------------------------------------------------------
// Returns a specific segment
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.getSegment = function(index)
{
    if (this.points.length < 2)
        return null;

    return new beat.toolchains.common.Segment(this.points[index], this.points[index + 1]);
}


//----------------------------------------------------------------------------------------
// Add a point at the end of the connection (only valid during the creation of a new
// connection)
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.addPoint = function(point, zoom)
{
    // Update the current active segment (if any)
    if (this.active_segment !== null)
    {
        this.active_segment.point1.toCells(zoom);

        if (this.active_segment.isVertical())
            this.active_segment.point2.x = beat.toolchains.common.cellsToPixels(this.active_segment.point1.col, zoom);
        else
            this.active_segment.point2.y = beat.toolchains.common.cellsToPixels(this.active_segment.point1.row, zoom);

        // Check if the previous segment must be removed
        if (this.active_segment_index > 0)
            this._cleanup();

        this.active_segment = null;
        this.active_segment_index = null;
    }

    // Store the new point in our list (in [pixels])
    if (point.vertical === undefined)
        point.vertical = !this.points[this.points.length - 1].vertical;

    point.toPixels(zoom);
    this.points.push(point);

    // Create the new active segment (if possible)
    if (this.points.length >= 2)
    {
        this.active_segment_index = this.points.length - 2;
        this.active_segment = new beat.toolchains.common.Segment(this.points[this.active_segment_index],
                                                                 this.points[this.active_segment_index + 1],
                                                                 true
                                                                );
    }
}


//----------------------------------------------------------------------------------------
// Finalize the creation of the connection
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.finalize = function(point, is_input, zoom)
{
    if (this.active_segment.isVertical())
    {
        this.active_segment.point1.toCells(zoom);
        this.active_segment.point1.col = point.col + (is_input ? -1 : 1);

        this.active_segment.point2.in_cells = true;
        this.active_segment.point2.col      = this.active_segment.point1.col;
        this.active_segment.point2.row      = point.row;
        this.active_segment.point2.x        = null;
        this.active_segment.point2.y        = null;
        this.active_segment.point2.fixed    = true;

        if (this.active_segment.isNull())
        {
            this.points.splice(this.active_segment_index + 1, 1);
            this.active_segment.point1.fixed = true;
        }

        this.points.push(point);
    }
    else
    {
        this.active_segment.point1.toCells(zoom);

        this.active_segment.point2.in_cells = true;
        this.active_segment.point2.col      = point.col + (is_input ? -1 : 1);
        this.active_segment.point2.row      = point.row;
        this.active_segment.point2.x        = null;
        this.active_segment.point2.y        = null;
        this.active_segment.point2.fixed    = true;

        if (this.active_segment.isNull())
        {
            this.points.splice(this.active_segment_index + 1, 1);
            this.active_segment.point1.fixed = true;
        }

        this.points.push(point);
    }

    this.active_segment = null;
    this.active_segment_index = null;

    this._cleanup();

    if (!is_input)
    {
        for (var i = 0; i < this.points.length - 1; ++i)
            this.points[i].vertical = this.points[i+1].vertical;

        this.points[this.points.length - 1].vertical = false;

        this.points.reverse();
    }
}


//----------------------------------------------------------------------------------------
// Update the connection using the provided mouse coordinates (only valid during the
// creation of a new connection)
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.update = function(x, y, zoom)
{
    if (this.active_segment !== null)
        this.active_segment.update(x, y, zoom);
}


//----------------------------------------------------------------------------------------
// Start a translation of the whole connection
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.startTranslation = function(zoom)
{
    if (this.selected == beat.toolchains.common.SELECTION_FULL)
    {
        for (var i = 0; i < this.points.length; ++i)
            this.points[i].toPixels(zoom);
    }

    else if (this.selected == beat.toolchains.common.SELECTION_INPUT)
    {
        var nb_segments = this.nbSegments();
        var vertical_segment_created = false;

        var segment = this.getSegment(nb_segments - 2);

        if (segment.isVertical())
        {
            // Create a (for the moment) 0-length horizontal segment
            var new_point = new beat.toolchains.common.Point(
                                        true, segment.point2.col, segment.point2.row, false
                                    );

            this.points.splice(nb_segments, 0, new_point);

            ++nb_segments;
        }
        else if (nb_segments == 3)
        {
            // Create a (for the moment) 0-length vertical segment
            var new_point1 = new beat.toolchains.common.Point(
                                        true, Math.floor((segment.point2.col + segment.point1.col) / 2),
                                        segment.point2.row, false
                                    );

            var new_point2 = new beat.toolchains.common.Point(
                                        true, new_point1.col, new_point1.row, true
                                    );

            this.points.splice(2, 0, new_point1, new_point2);

            nb_segments += 2;

            vertical_segment_created = true;
        }

        this.translation_data = {
            input_segment:            this.getSegment(nb_segments - 1),
            horizontal_segment:       this.getSegment(nb_segments - 2),
            vertical_segment:         this.getSegment(nb_segments - 3),
            vertical_segment_created: vertical_segment_created,
        };

        this.translation_data.input_segment.point2.toPixels(zoom);
        this.translation_data.horizontal_segment.point2.toPixels(zoom);
        this.translation_data.vertical_segment.point2.toPixels(zoom);
    }

    else if (this.selected == beat.toolchains.common.SELECTION_OUTPUT)
    {
        var vertical_segment_created = false;

        var segment = this.getSegment(1);

        if (segment.isVertical())
        {
            // Create a (for the moment) 0-length horizontal segment
            var new_point = new beat.toolchains.common.Point(
                                        true, segment.point1.col, segment.point1.row, false
                                    );

            this.points.splice(1, 0, new_point);
        }
        else if (this.nbSegments() == 3)
        {
            // Create a (for the moment) 0-length vertical segment
            var new_point1 = new beat.toolchains.common.Point(
                                        true, Math.floor((segment.point2.col + segment.point1.col) / 2),
                                        segment.point2.row, false
                                    );

            var new_point2 = new beat.toolchains.common.Point(
                                        true, new_point1.col, new_point1.row, true
                                    );

            this.points.splice(2, 0, new_point1, new_point2);

            vertical_segment_created = true;
        }

        this.translation_data = {
            output_segment:           this.getSegment(0),
            horizontal_segment:       this.getSegment(1),
            vertical_segment:         this.getSegment(2),
            vertical_segment_created: vertical_segment_created,
        };

        this.translation_data.output_segment.point1.toPixels(zoom);
        this.translation_data.horizontal_segment.point1.toPixels(zoom);
        this.translation_data.vertical_segment.point1.toPixels(zoom);
    }
}


//----------------------------------------------------------------------------------------
// End a translation of the whole connection
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.endTranslation = function(zoom)
{
    if (this.selected == beat.toolchains.common.SELECTION_FULL)
    {
        if (this.nbSegments() == 3)
        {
            var segment = this.getSegment(0);

            segment.point1.col = this.connection.output.block.representation.left + this.connection.output.block.representation.width;
            segment.point1.row = this.connection.output.block.representation.top + 2 + this.connection.output.index;
            segment.point1.in_cells = true;

            segment.point2.col = segment.point1.col + 1;
            segment.point2.row = segment.point1.row;
            segment.point2.in_cells = true;

            var segment2 = this.getSegment(2);

            segment2.point2.col = this.connection.input.block.representation.left;
            segment2.point2.row = segment.point1.row;
            segment2.point2.in_cells = true;

            segment2.point1.col = segment2.point2.col - 1;
            segment2.point1.row = segment.point1.row;
            segment2.point1.in_cells = true;
        }
        else
        {
            var segment = this.getSegment(0);

            var dx = beat.toolchains.common.cellsToPixels(this.connection.output.block.representation.left + this.connection.output.block.representation.width, zoom) - segment.point1.x;
            var dy = beat.toolchains.common.cellsToPixels(this.connection.output.block.representation.top + 2 + this.connection.output.index, zoom) - segment.point1.y;

            for (var i = 0; i < this.points.length; ++i)
            {
                var point = this.points[i];

                point.x += dx;
                point.y += dy;
                point.toCells(zoom);
            }
        }
    }

    else if (this.selected == beat.toolchains.common.SELECTION_INPUT)
    {
        this.translation_data.input_segment.point2.col = this.connection.input.block.representation.left;
        this.translation_data.input_segment.point2.row = this.connection.input.block.representation.top + 2 + this.connection.input.index;
        this.translation_data.input_segment.point2.in_cells = true;

        this.translation_data.horizontal_segment.point2.col = this.translation_data.input_segment.point2.col - 1;
        this.translation_data.horizontal_segment.point2.row = this.translation_data.input_segment.point2.row;
        this.translation_data.horizontal_segment.point2.in_cells = true;

        this.translation_data.vertical_segment.point2.col = this.translation_data.vertical_segment.point1.col;
        this.translation_data.vertical_segment.point2.row = this.translation_data.horizontal_segment.point2.row;
        this.translation_data.vertical_segment.point2.in_cells = true;

        this.translation_data = null;
    }

    else if (this.selected == beat.toolchains.common.SELECTION_OUTPUT)
    {
        this.translation_data.output_segment.point1.col = this.connection.output.block.representation.left + this.connection.output.block.representation.width;
        this.translation_data.output_segment.point1.row = this.connection.output.block.representation.top + 2 + this.connection.output.index;
        this.translation_data.output_segment.point1.in_cells = true;

        this.translation_data.horizontal_segment.point1.col = this.translation_data.output_segment.point1.col + 1;
        this.translation_data.horizontal_segment.point1.row = this.translation_data.output_segment.point1.row;
        this.translation_data.horizontal_segment.point1.in_cells = true;

        this.translation_data.vertical_segment.point1.col = this.translation_data.vertical_segment.point2.col;
        this.translation_data.vertical_segment.point1.row = this.translation_data.horizontal_segment.point1.row;
        this.translation_data.vertical_segment.point1.in_cells = true;

        this.translation_data = null;
    }

    this._cleanup();
}


//----------------------------------------------------------------------------------------
// Move the whole connection to the given coordinates [in pixels]
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.translate = function(dx, dy, zoom)
{
    if (this.selected == beat.toolchains.common.SELECTION_FULL)
    {
        for (var i = 0; i < this.points.length; ++i)
        {
            this.points[i].x += dx;
            this.points[i].y += dy;
        }
    }

    else if (this.selected == beat.toolchains.common.SELECTION_INPUT)
    {
        this.translation_data.input_segment.point2.x += dx;
        this.translation_data.input_segment.point2.y += dy;

        this.translation_data.horizontal_segment.point2.x = this.translation_data.input_segment.point2.x - beat.toolchains.common.cellsToPixels(1, zoom);
        this.translation_data.horizontal_segment.point2.y = this.translation_data.input_segment.point2.y;

        this.translation_data.vertical_segment.point2.y += dy;

        if (this.translation_data.vertical_segment_created &&
            (beat.toolchains.common.cellsToPixels(this.translation_data.vertical_segment.point1.row, zoom) == this.translation_data.vertical_segment.point2.y))
        {
            var previous_segment = this.getSegment(this.nbSegments() - 4);

            this.translation_data.vertical_segment.point1.col = Math.floor((previous_segment.point1.col +
                                                                            beat.toolchains.common.pixelsToCells(this.translation_data.horizontal_segment.point2.x, zoom)) / 2);

            this.translation_data.vertical_segment.point2.x =
                    beat.toolchains.common.cellsToPixels(this.translation_data.vertical_segment.point1.col, zoom);
        }
    }

    else if (this.selected == beat.toolchains.common.SELECTION_OUTPUT)
    {
        this.translation_data.output_segment.point1.x += dx;
        this.translation_data.output_segment.point1.y += dy;

        this.translation_data.horizontal_segment.point1.x = this.translation_data.output_segment.point1.x + beat.toolchains.common.cellsToPixels(1, zoom);
        this.translation_data.horizontal_segment.point1.y = this.translation_data.output_segment.point1.y;

        this.translation_data.vertical_segment.point1.y += dy;

        if (this.translation_data.vertical_segment_created &&
            (beat.toolchains.common.cellsToPixels(this.translation_data.vertical_segment.point2.row, zoom) == this.translation_data.vertical_segment.point1.y))
        {
            var next_segment = this.getSegment(3);

            this.translation_data.vertical_segment.point2.col = Math.floor((next_segment.point2.col +
                                                                            beat.toolchains.common.pixelsToCells(this.translation_data.horizontal_segment.point1.x, zoom)) / 2);

            this.translation_data.vertical_segment.point1.x =
                    beat.toolchains.common.cellsToPixels(this.translation_data.vertical_segment.point2.col, zoom);
        }
    }
}


//----------------------------------------------------------------------------------------
// Start a translation of a single segment of the connection
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.startSegmentTranslation = function(segment_index, zoom)
{
    var segment = this.getSegment(segment_index);

    if (segment.isVertical())
    {
        if (segment_index == this.nbSegments() - 2)
        {
            // Create a (for the moment) 0-length horizontal segment
            var new_point = new beat.toolchains.common.Point(
                                        true, segment.point2.col, segment.point2.row, false
                                    );

            this.points.splice(segment_index + 2, 0, new_point);
        }

        if (segment_index == 1)
        {
            // Create a (for the moment) 0-length horizontal segment
            var new_point = new beat.toolchains.common.Point(
                                        true, segment.point1.col, segment.point1.row, false
                                    );

            this.points.splice(1, 0, new_point);

            ++segment_index;
            segment = this.getSegment(segment_index);
        }
    }
    else
    {
        if (segment_index == this.nbSegments() - 2)
        {
            // Create a (for the moment) 0-length vertical segment
            var new_point = new beat.toolchains.common.Point(
                                        true, segment.point2.col, segment.point2.row, true
                                    );

            this.points.splice(segment_index + 2, 0, new_point);
        }

        if (segment_index == 1)
        {
            // Create a (for the moment) 0-length vertical segment
            var new_point = new beat.toolchains.common.Point(
                                        true, segment.point1.col, segment.point1.row, true
                                    );

            this.points.splice(2, 0, new_point);

            ++segment_index;
            segment = this.getSegment(segment_index);
        }
    }

    this.translation_data = {
        segment_index: segment_index,
        original_col:  segment.point1.col,
        original_row:  segment.point1.row,
    };

    segment.point1.toPixels(zoom);
    segment.point2.toPixels(zoom);

    return segment_index;
}


//----------------------------------------------------------------------------------------
// Translate a single segment of the connection
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.translateSegment = function(diff)
{
    var segment = this.getSegment(this.translation_data.segment_index);

    if (segment.isVertical())
    {
        segment.point1.x += diff;
        segment.point2.x += diff;
    }
    else
    {
        segment.point1.y += diff;
        segment.point2.y += diff;
    }
}


//----------------------------------------------------------------------------------------
// Terminate the translation of a single segment of the connection
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.endSegmentTranslation = function(zoom)
{
    var segment = this.getSegment(this.translation_data.segment_index);

    segment.point1.toCells(zoom);
    segment.point2.toCells(zoom);

    var new_segment_index = this.translation_data.segment_index;

    var previous_segment = this.getSegment(this.translation_data.segment_index - 1);

    if (previous_segment.isNull())
    {
        var previous_segment2 = this.getSegment(this.translation_data.segment_index - 2);

        if (segment.isVertical())
        {
            if (previous_segment2.isVertical())
                new_segment_index = this.translation_data.segment_index - 2;
            else
                new_segment_index = this.translation_data.segment_index - 1;
        }
        else
        {
            if (!previous_segment2.isVertical() && (this.translation_data.segment_index > 2))
                new_segment_index = this.translation_data.segment_index - 2;
            else
                new_segment_index = this.translation_data.segment_index - 1;
        }
    }

    this._cleanup();

    this.translation_data = null;

    return new_segment_index;
}


//----------------------------------------------------------------------------------------
// Terminate the translation of a single segment of the connection
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.cancelSegmentTranslation = function(zoom)
{
    var segment = this.getSegment(this.translation_data.segment_index);

    segment.point1.toCells(zoom);
    segment.point2.toCells(zoom);

    if (segment.isVertical())
    {
        segment.point1.col = this.translation_data.original_col;
        segment.point2.col = this.translation_data.original_col;
    }
    else
    {
        segment.point1.row = this.translation_data.original_row;
        segment.point2.row = this.translation_data.original_row;
    }

    this._cleanup();

    this.translation_data = null;
}


//----------------------------------------------------------------------------------------
// Rescale the whole connection (to use when the zoom changes during a
// translation/creation)
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.scale = function(scale, zoom)
{
    for (var i = 0; i < this.points.length; ++i)
    {
        if (!this.points[i].in_cells)
        {
            if ((i > 0) && this.points[i-1].in_cells)
            {
                if (this.points[i].vertical)
                {
                    this.points[i].x = beat.toolchains.common.cellsToPixels(this.points[i-1].col, zoom);
                    this.points[i].y *= scale;
                }
                else
                {
                    this.points[i].x *= scale;
                    this.points[i].y = beat.toolchains.common.cellsToPixels(this.points[i-1].row, zoom);
                }
            }
            else
            {
                this.points[i].x *= scale;
                this.points[i].y *= scale;
            }
        }
    }
}


//----------------------------------------------------------------------------------------
// Split a horizontal segment in half (create a vertical one in the middle)
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype.splitHorizontalSegment = function(index, zoom)
{
    var segment = this.getSegment(index);

    var new_point1 = new beat.toolchains.common.Point(
                                false, (segment.point2.x + segment.point1.x) / 2,
                                segment.point2.y, false
                            );

    var new_point2 = new_point1.toPixelsCopy(zoom);
    new_point2.vertical = true;

    this.points.splice(index + 1, 0, new_point1, new_point2);
}


//----------------------------------------------------------------------------------------
// Remove superfluous (0-length) segments
//----------------------------------------------------------------------------------------
beat.toolchains.common.ConnectionRepresentation.prototype._cleanup = function()
{
    var segment_index = 0;

    while (segment_index < this.points.length - 1)
    {
        var segment = this.getSegment(segment_index);

        if (!segment.isFixed() && segment.isNull())
        {
            if (segment.point2.fixed)
                segment.point1.fixed = true;

            this.points.splice(segment_index + 1, 1);
            segment_index = 0;
            continue;
        }

        else if (!segment.isFixed() && (segment_index > 0))
        {
            var previous_segment = this.getSegment(segment_index - 1);

            if (!previous_segment.isFixed() &&
                (previous_segment.isVertical() == segment.isVertical()))
            {
                this.points.splice(segment_index, 1);
                segment_index = 0;
                continue;
            }
        }

        ++segment_index;
    }


    if (this.active_segment === null)
    {
        var first_segment = this.getSegment(0);
        if (first_segment.getLength() > 1)
        {
            var new_point = new beat.toolchains.common.Point(true, first_segment.point1.col + 1,
                                                             first_segment.point1.row, false, true);

            first_segment.point2.fixed = false;

            this.points.splice(1, 0, new_point);
        }


        var last_segment = this.getSegment(this.nbSegments() - 1);
        if (last_segment.getLength() > 1)
        {
            var new_point = new beat.toolchains.common.Point(true, last_segment.point2.col - 1,
                                                             last_segment.point2.row, false, true);

            last_segment.point1.fixed = false;

            this.points.splice(this.points.length - 1, 0, new_point);
        }
    }


    this.bounding_box = {
        left:   this.points[0].col,
        top:    this.points[0].row,
        right:  this.points[0].col,
        bottom: this.points[0].row,
    };

    for (var i = 1; i < this.points.length; ++i)
    {
        var point = this.points[i];

        if (point.col < this.bounding_box.left)
            this.bounding_box.left = point.col;
        else if (point.col > this.bounding_box.right)
            this.bounding_box.right = point.col;

        if (point.row < this.bounding_box.top)
            this.bounding_box.top = point.row;
        else if (point.row > this.bounding_box.bottom)
            this.bounding_box.bottom = point.row;
    }
}



/************************************ CLASS: Point **************************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.common.Point = function(in_cells, x_or_col, y_or_row, vertical, fixed)
{
    this.in_cells = in_cells;
    this.col      = (in_cells ? x_or_col : null);
    this.row      = (in_cells ? y_or_row : null);
    this.x        = (!in_cells ? x_or_col : null);
    this.y        = (!in_cells ? y_or_row : null);
    this.vertical = vertical;
    this.fixed    = (fixed === true);
}


//----------------------------------------------------------------------------------------
// Convert the coordinates of the point in [pixels]
//----------------------------------------------------------------------------------------
beat.toolchains.common.Point.prototype.toPixels = function(zoom)
{
    if (!this.in_cells)
        return;

    this.x        = beat.toolchains.common.cellsToPixels(this.col, zoom);
    this.y        = beat.toolchains.common.cellsToPixels(this.row, zoom);
    this.col      = null;
    this.row      = null;
    this.in_cells = false;
}


//----------------------------------------------------------------------------------------
// Returns a copy of the point, with the coordinates in [pixels]
//----------------------------------------------------------------------------------------
beat.toolchains.common.Point.prototype.toPixelsCopy = function(zoom)
{
    if (!this.in_cells)
    {
        return new beat.toolchains.common.Point(false, this.x, this.y, this.vertical, this.fixed);
    }
    else
    {
        return new beat.toolchains.common.Point(false, beat.toolchains.common.cellsToPixels(this.col, zoom),
                                                beat.toolchains.common.cellsToPixels(this.row, zoom),
                                                this.vertical, this.fixed);
    }
}


//----------------------------------------------------------------------------------------
// Convert the coordinates of the point in [cells] (snap to the nearest cell)
//----------------------------------------------------------------------------------------
beat.toolchains.common.Point.prototype.toCells = function(zoom)
{
    if (this.in_cells)
        return;

    this.col      = beat.toolchains.common.nearestCell(this.x, zoom);
    this.row      = beat.toolchains.common.nearestCell(this.y, zoom);
    this.x        = null;
    this.y        = null;
    this.in_cells = true;
}


//----------------------------------------------------------------------------------------
// Convert the coordinates of the point in [cells] (snap to the lower cell)
//----------------------------------------------------------------------------------------
beat.toolchains.common.Point.prototype.toCells2 = function(zoom)
{
    if (this.in_cells)
        return;

    this.col      = beat.toolchains.common.pixelsToCells(this.x, zoom);
    this.row      = beat.toolchains.common.pixelsToCells(this.y, zoom);
    this.x        = null;
    this.y        = null;
    this.in_cells = true;
}



/********************************** CLASS: CellPoint ************************************/

//----------------------------------------------------------------------------------------
// Constructor
//----------------------------------------------------------------------------------------
beat.toolchains.common.Segment = function(point1, point2, can_modify_length)
{
    this.point1            = point1;
    this.point2            = point2;
    this.can_modify_length = (can_modify_length === true);
}


//----------------------------------------------------------------------------------------
// Indicates if the segment is vertical
//----------------------------------------------------------------------------------------
beat.toolchains.common.Segment.prototype.isVertical = function()
{
    return this.point2.vertical;
}


//----------------------------------------------------------------------------------------
// Indicates if the segment is vertical
//----------------------------------------------------------------------------------------
beat.toolchains.common.Segment.prototype.isFixed = function()
{
    return (this.point1.fixed && this.point2.fixed);
}


//----------------------------------------------------------------------------------------
// Indicates if the segment is null (0-length). Only when point coordinates are in [cells]
//----------------------------------------------------------------------------------------
beat.toolchains.common.Segment.prototype.isNull = function()
{
    return (this.point1.in_cells && this.point2.in_cells &&
           (this.point1.col == this.point2.col) && (this.point1.row == this.point2.row));
}


//----------------------------------------------------------------------------------------
// Returns the length of the segment. Only when point coordinates are in [cells]
//----------------------------------------------------------------------------------------
beat.toolchains.common.Segment.prototype.getLength = function()
{
    if (this.point2.vertical)
        return Math.abs(this.point2.row - this.point1.row);
    else
        return Math.abs(this.point2.col - this.point1.col);
}


//----------------------------------------------------------------------------------------
// Update the connection using the provided mouse coordinates
//----------------------------------------------------------------------------------------
beat.toolchains.common.Segment.prototype.update = function(x, y, zoom)
{
    this.point1.toPixels(zoom);
    this.point2.toPixels(zoom);

    if (this.isVertical())
    {
        if (this.can_modify_length)
            this.point2.y = y;

        this.point1.x = x;
        this.point2.x = x;
    }
    else
    {
        if (this.can_modify_length)
            this.point2.x = x;

        this.point1.y = y;
        this.point2.y = y;
    }
}
